import base64
import io
import json
import random
from itertools import accumulate
from typing import Optional, Any
from zipfile import ZipFile, ZIP_DEFLATED

import yaml

from Class import settingkey
from Class.exceptions import GeneratorException
from Class.itemClass import ItemEncoder
from Class.newLocationClass import KH2Location
from Class.openkhmod import ATKPObject, AttackEntriesOrganizer, ModYml
from Class.seedSettings import SeedSettings, ExtraConfigurationData, makeKHBRSettings
from List import ChestList
from List.DropRateIds import id_to_enemy_name
from List.ItemList import Items
from List.LvupStats import DreamWeaponOffsets
from List.ObjectiveList import KH2Objective
from List.configDict import itemType, locationCategory, locationType, BattleLevelOption
from List.inventory import bonus, misc
from List.location import simulatedtwilighttown as stt
from Module import hashimage
from Module.RandomizerSettings import RandomizerSettings
from Module.battleLevels import BtlvViewer
from Module.cosmetics import CosmeticsMod
from Module.hints import Hints, HintData
from Module.knockbackTypes import KnockbackTypes
from Module.multiworld import MultiWorldOutput
from Module.newRandomize import Randomizer, SynthesisRecipe, ItemAssignment
from Module.resources import resource_path
from Module.seedmod import SeedModBuilder, ChestVisualAssignment, CosmeticsModAppender
from Module.spoilerLog import (
    item_spoiler_dictionary,
    levelStatsDictionary,
    synth_recipe_dictionary,
    weapon_stats_dictionary,
    objectives_dictionary,
)
from Module.version import LOCAL_UI_VERSION


def noop(self, *args, **kw):
    pass


def number_to_bytes(item) -> tuple[int, int]:
    # for byte1, find the most significant bits from the item Id
    item_byte1 = item >> 8
    # for byte0, isolate the least significant bits from the item Id
    item_byte0 = item & 0x00FF
    return item_byte0, item_byte1


def bytes_to_number(byte0, byte1=0) -> int:
    return int(byte0) + int(byte1 << 8)


def _assignment_subset(
    assigned: list[ItemAssignment], categories: list[locationCategory]
) -> list[ItemAssignment]:
    return [
        assignment
        for assignment in assigned
        if any(item is assignment.location.LocationCategory for item in categories)
    ]


def _assignment_subset_from_type(
    assigned: list[ItemAssignment], types: list[locationType]
) -> list[ItemAssignment]:
    return [
        assignment
        for assignment in assigned
        if any(item in assignment.location.LocationTypes for item in types)
    ]


def _run_khbr(
    platform: str, enemy_options: dict, mod_yml: ModYml, out_zip: ZipFile
) -> tuple[Optional[str], dict[str, list]]:
    if platform == "PC":
        enemy_options["memory_expansion"] = True
    else:
        enemy_options["memory_expansion"] = False

    enemy_spoilers = _invoke_khbr_with_overrides(enemy_options, mod_yml.data, out_zip)

    lines = enemy_spoilers.split("\n")

    current_key = ""
    enemy_spoilers_json = {}
    for line in lines:
        if "\t" in line:
            modded_line = line.replace("\t", "")
            enemies = modded_line.split(" became ")
            # this is adding to the current list
            new_entry = {"original": enemies[0], "new": enemies[1]}
            enemy_spoilers_json[current_key].append(new_entry)
        elif line != "":
            current_key = line
            enemy_spoilers_json[current_key] = []
    if enemy_spoilers_json:
        out_zip.writestr(
            "enemies.rando",
            base64.b64encode(json.dumps(enemy_spoilers_json).encode("utf-8")).decode(
                "utf-8"
            ),
        )
        # for boss_replacement in enemy_spoilers_json["BOSSES"]:
        #     print(f"{boss_replacement['original']} {boss_replacement['new']}")

    return enemy_spoilers, enemy_spoilers_json


def _add_cosmetics(out_zip: ZipFile, mod_yml: ModYml, settings: SeedSettings):
    appender = CosmeticsModAppender(out_zip=out_zip, mod_yml=mod_yml)

    mod_yml.add_assets(CosmeticsMod.randomize_field2d(settings))
    mod_yml.add_assets(CosmeticsMod.randomize_itempics(settings))
    mod_yml.add_assets(CosmeticsMod.randomize_end_screen(settings))

    keyblade_assets, keyblade_replacements = CosmeticsMod.randomize_keyblades(settings)
    appender.write_keyblade_rando_assets(keyblade_assets, keyblade_replacements)

    music_assets, music_replacements = CosmeticsMod.randomize_music(settings)
    appender.write_music_rando_assets(music_assets, music_replacements)

    from Module.cosmeticsmods.texture import TextureRecolorizer
    texture_assets = TextureRecolorizer(settings).recolor_textures()
    mod_yml.add_assets(texture_assets)

    if settings.get(settingkey.RANDO_THEMED_TEXTURES):
        appender.write_rando_themed_texture_assets()


class SynthList:
    def __init__(self, offset: int, binary: bytearray):
        self.offset = offset
        self.id = bytes_to_number(binary[0], binary[1])
        self.reward = bytes_to_number(binary[2], binary[3])
        self.reward_type = bytes_to_number(binary[4])
        self.material_type = bytes_to_number(binary[5])
        self.material_rank = bytes_to_number(binary[6])
        self.condition_type = bytes_to_number(binary[7])
        self.count_needed = bytes_to_number(binary[8], binary[9])
        self.unlock_event_shop = bytes_to_number(binary[10], binary[11])

    def __str__(self):
        if self.reward_type == 1:
            return f"{self.offset}: ID{self.id} R{self.reward} RT{self.reward_type} MT{self.material_type} MR{self.material_rank} CT{self.condition_type} CN{self.count_needed}"
        else:
            return ""


class DropRates:
    def __init__(self, offset: int, binary: bytearray):
        self.offset = offset
        self.id = bytes_to_number(binary[0], binary[1])
        self.small_hp = bytes_to_number(binary[2])
        self.big_hp = bytes_to_number(binary[3])
        self.big_munny = bytes_to_number(binary[4])
        self.medium_munny = bytes_to_number(binary[5])
        self.small_munny = bytes_to_number(binary[6])
        self.small_mp = bytes_to_number(binary[7])
        self.big_mp = bytes_to_number(binary[8])
        self.small_drive = bytes_to_number(binary[9])
        self.big_drive = bytes_to_number(binary[10])
        self.item1 = bytes_to_number(binary[12], binary[13])
        self.item1_chance = bytes_to_number(binary[14], binary[15])
        self.item2 = bytes_to_number(binary[16], binary[17])
        self.item2_chance = bytes_to_number(binary[18], binary[19])
        self.item3 = bytes_to_number(binary[20], binary[21])
        self.item3_chance = bytes_to_number(binary[22], binary[23])

    def __str__(self):
        if True:  # self.item1_chance and self.id not in id_to_enemy_name:
            dummy = ""
            return f"{self.id} {(id_to_enemy_name[self.id] if self.id in id_to_enemy_name else dummy)} \n HP ({self.small_hp},{self.big_hp}) \n MP ({self.small_mp},{self.big_mp}) \n Munny ({self.small_munny},{self.medium_munny},{self.big_munny}) \n Orbs ({self.small_drive},{self.big_drive}) Items: \n--- {self.item1} ({self.item1_chance/1.0}%)\n--- {self.item2} ({self.item2_chance/1.0}%)\n--- {self.item3} ({self.item3_chance/1.0}%)"
        # elif self.id in id_to_enemy_name:
        #     return f"{self.id} {id_to_enemy_name[self.id]}"
        # else:
        #     return ""


class SynthLocation:
    def __init__(self, loc: int, item: int, in_recipe: SynthesisRecipe):
        self.location = loc
        self.item = item
        self.requirements = [(0, 0)] * 6
        self.unlock_rank = in_recipe.unlock_rank
        for i, recipe_requirement in enumerate(in_recipe.requirements):
            self.requirements[i] = (
                recipe_requirement.synth_item.Id,
                recipe_requirement.amount,
            )

    def get_starting_location(self) -> int:
        # header bytes + offset to the specific recipe + skip over the recipe bytes
        return 16 + self.location * 32 + 2

    def get_bytes(self) -> list[int]:
        binary = [self.unlock_rank, 0]  # unlock condition/rank
        item_byte0, item_byte1 = number_to_bytes(self.item)
        # add the item for this recipe
        binary.append(item_byte0)
        binary.append(item_byte1)
        # add the item as the upgraded version
        binary.append(item_byte0)
        binary.append(item_byte1)

        for req in self.requirements:
            item_byte0, item_byte1 = number_to_bytes(req[0])
            # add the item as an ingredient
            binary.append(item_byte0)
            binary.append(item_byte1)
            item_byte0, item_byte1 = number_to_bytes(req[1])
            # add the amount of that ingredient
            binary.append(item_byte0)
            binary.append(item_byte1)
        return binary


# (output zip, spoiler log, enemy log)
SeedZipResult = tuple[io.BytesIO, Optional[str], Optional[str]]
SeedNotZipResult = tuple[Optional[str], Optional[str]]


def _invoke_khbr_with_overrides(
    enemy_options: dict, mod: dict[str, Any], out_zip: ZipFile
):
    # """
    # A function that mimics the call to generateToZip in khbr. Splitting this out to allow for granular control of
    # khbr's boss placement.
    # """
    # from khbr.randomizer import Randomizer as BossEnemyRandomizer
    # from khbr.KH2.EnemyManager import EnemyManager
    # from khbr.textutils import create_spoiler_text

    random.seed(
        str(enemy_options)
    )  # seed boss/enemy with enemy options, which should include the seed name
    del enemy_options["seed_name"]  # remove it so khbr doesn't complain

    # khbr_randomizer = BossEnemyRandomizer()

    # # get kh2's data
    # game_data = khbr_randomizer._get_game("kh2")

    # # make a new enemy manager with custom jsons
    # game_data.enemy_manager = EnemyManager(resource_path("static/khbr_override/"))

    # # ###  MANUAL STEP: updating records
    # # import os
    # # full_records = game_data.enemy_manager.create_enemy_records(ispc=False)
    # # name = "full_enemy_records.json"
    # # json.dump(full_records, open(os.path.join(resource_path("static/khbr_override/"), name), "w"), indent=4)

    # # full_records = game_data.enemy_manager.create_enemy_records(ispc=True)
    # # name = "full_enemy_records_pc.json"
    # # json.dump(full_records, open(os.path.join(resource_path("static/khbr_override/"), name), "w"), indent=4)

    # # ### DEBUGGING
    # # valid_boss_replacements = {}
    # # boss_list = game_data.enemy_manager.get_boss_list(enemy_options)
    # # for _,source_boss in boss_list.items():
    # #     valid_boss_replacements[source_boss["name"]] = []
    # # for _,source_boss in boss_list.items():
    # #     for __,dest_boss in boss_list.items():
    # #         if source_boss["name"]!=dest_boss["name"]:
    # #             if not EnemyManager.isReplacementBlocked(source_boss,dest_boss):
    # #                 valid_boss_replacements[source_boss["name"]].append(dest_boss["name"])
    # # with open("valid_boss_replacements.json","w") as f:
    # #     json.dump(valid_boss_replacements, f, indent=4)

    # # make sure the options are formatted properly
    # khbr_randomizer._validate_options(
    #     khbr_randomizer.get_options_cli(game_data), enemy_options
    # )

    # # actually perform the randomization
    # randomization = game_data.perform_randomization(enemy_options)

    # # add the assets to the zip and mod yml
    # assets = game_data.generate_files(randomization=randomization, outzip=out_zip)
    # mod["assets"].extend(assets)

    # # create the spoiler log
    # return create_spoiler_text(game_data.spoilers)

    ### Backup old way
    from khbr.randomizer import Randomizer as BossEnemyRandomizer
    enemySpoilers = BossEnemyRandomizer().generateToZip("kh2", enemy_options, mod, out_zip)
    return enemySpoilers


class SeedZip:
    def __init__(
        self,
        settings: RandomizerSettings,
        randomizer: Randomizer,
        hints: HintData,
        extra_data: ExtraConfigurationData,
        location_spheres: dict[KH2Location,int],
        multiworld: Optional[MultiWorldOutput] = None,
    ):
        self.settings = settings
        self.randomizer = randomizer
        self.hints = hints
        self.extra_data = extra_data
        self.location_spheres = location_spheres
        self.multiworld = multiworld

    def make_spoiler_without_zip(self) -> str:
        enemy_spoilers_json: dict[str, str] = {}

        settings = self.settings
        btlv_option_name = settings.battle_level_rando

        btlv = BtlvViewer()
        btlv.use_setting(
            btlv_option_name,
            battle_level_offset=settings.battle_level_offset,
            battle_level_range=settings.battle_level_range,
            battle_level_random_min_max=settings.battle_level_random_min_max,
        )
        battle_level_spoiler = btlv.get_spoiler()
        journal_hints_spoiler = {}
        spoiler_log_output = self.generate_spoiler_html(
            enemy_spoilers_json, battle_level_spoiler, journal_hints_spoiler
        )
        return spoiler_log_output

    def create_zip(self) -> SeedZipResult:
        settings = self.settings
        spoiler_log = settings.spoiler_log
        extra_data = self.extra_data
        tourney_gen = extra_data.tourney

        title = "Randomizer Seed"
        if spoiler_log and not tourney_gen:
            title += " w/ Spoiler"
        title+=" "+LOCAL_UI_VERSION

        zip_data = io.BytesIO()
        spoiler_log_output: Optional[str] = None
        enemy_log_output: Optional[str] = None
        with ZipFile(zip_data, "w", ZIP_DEFLATED) as out_zip:
            yaml.emitter.Emitter.process_tag = noop
            mod = SeedModBuilder(title, out_zip)
            mod.add_base_assets()
            mod.add_base_messages(
                seed_hash_icons=settings.seedHashIcons,
                crit_mode=settings.crit_mode,
                final_door_requirement_message=self._final_door_requirement_message()
            )
            self.prepare_companion_damage_knockback(mod)

            if settings.dummy_forms:
                # convert the valor and final ids to their dummy values
                for a in self.randomizer.assignments:
                    for dummy_form_item in Items.getDummyFormItems():
                        if a.item.Name == dummy_form_item.Name:
                            a.item = dummy_form_item
                # if valor/final in starting inventory, swap their ids
                orig_to_dummy = Items.getFormToDummyMap()
                for orig_id,dummy_id in orig_to_dummy.items():
                    self.randomizer.starting_item_ids[:] = [dummy_id if x==orig_id else x for x in self.randomizer.starting_item_ids]

            self.assign_treasures(mod)
            self.assign_levels(mod)
            self.assign_bonuses(mod)
            self.assign_form_levels(mod)
            self.assign_weapon_stats(mod)
            self.assign_starting_items(mod)

            enemy_spoilers, enemy_spoilers_json = self.run_khbr_if_needed(mod, out_zip)

            if self.multiworld:
                out_zip.writestr("multiworld.multi", json.dumps(self.multiworld()))

            self.generate_seed_hash_image(out_zip)
            self.create_puzzle_assets(mod)
            self.create_synth_assets(mod)
            if settings.as_data_split:
                mod.write_as_data_split_assets()
            if settings.disable_antiform:
                self.disable_antiform(mod)
            if settings.skip_carpet_escape:
                mod.write_skip_carpet_escape_assets()
            if settings.pr_map_skip:
                mod.write_map_skip_assets()
            if settings.block_cor_skip:
                mod.write_block_cor_skip_assets()
            if settings.block_shan_yu_skip:
                mod.write_block_shan_yu_skip_assets()
            if settings.atlantica_skip:
                mod.write_atlantica_tutorial_skip_assets()
            if settings.wardrobe_skip:
                mod.write_wardrobe_skip_assets()
            self.create_drop_rate_assets(mod)
            self.create_shop_rando_assets(mod)
            if settings.objective_rando:
                self.create_objective_rando_assets(mod,settings.num_objectives_needed, self.randomizer.objectives)
            if settings.emblems:
                self.create_emblem_rando_assets(mod,settings.num_emblems_needed)
            self.create_chest_visual_assets(mod)
            battle_level_spoiler = self.create_battle_level_rando_assets(mod)
            journal_hints_spoiler: dict[str, str] = {}

            hints = self.hints
            if hints is not None:
                Hints.write_hints(hints, out_zip)
                Hints.write_hint_text(hints, mod, journal_hints_spoiler)

            if settings.roxas_abilities_enabled:
                boss_enabled = not settings.enemy_options.get("boss", False) in [
                    False,
                    "Disabled",
                ]
                mod.write_better_stt_assets(boss_enabled)
            
            if settings.keyblades_unlock_chests:
                mod.write_keyblade_locking_lua()

            # if "beta" in LOCAL_UI_VERSION:
            #     print("Writing beta stuff")
            #     mod.write_goa_lua()

            self.add_cmd_list_modifications(mod)

            if spoiler_log or tourney_gen:
                # For a tourney seed, generate the spoiler log to return to the caller but don't include it in the zip
                spoiler_log_output = self.generate_spoiler_html(
                    enemy_spoilers_json, battle_level_spoiler, journal_hints_spoiler
                )
                if not tourney_gen:
                    out_zip.writestr("spoilerlog.html", spoiler_log_output)
                    out_zip.write(resource_path("static/KHMenu.otf"), "misc/KHMenu.otf")

                # For a tourney seed, return the enemy log to the caller but don't include it in the zip
                enemy_log_output = enemy_spoilers
                if enemy_spoilers and not tourney_gen:
                    out_zip.writestr("enemyspoilers.txt", enemy_spoilers)

            _add_cosmetics(out_zip=out_zip, mod_yml=mod.mod_yml, settings=settings.ui_settings)

            out_zip.write(resource_path("Module/icon.png"), "icon.png")

            mod.write_mod_ymls(
                include_main_mod_yml=False
            )  # We'll add the main mod.yml after the validation
            out_zip.close()
        zip_data.seek(0)

        new_data = mod.validate_and_write_mod_yml(zip_data, settings)

        return new_data, spoiler_log_output, enemy_log_output

    def run_khbr_if_needed(
        self, mod: SeedModBuilder, out_zip: ZipFile
    ) -> tuple[Optional[str], dict[str, list]]:
        enemy_options = self.settings.enemy_options

        def _should_run_khbr() -> bool:
            if not enemy_options.get("boss", False) in [False, "Disabled"]:
                return True
            if not enemy_options.get("enemy", False) in [False, "Disabled"]:
                return True
            if enemy_options.get("remove_damage_cap", False):
                return True
            if enemy_options.get("cups_give_xp", False):
                return True
            if enemy_options.get("retry_data_final_xemnas", False):
                return True
            if enemy_options.get("retry_dark_thorn", False):
                return True
            if enemy_options.get("costume_rando", False):
                return True
            if enemy_options.get("party_member_rando", False):
                return True
            if not enemy_options.get("remove_cutscenes", False) in [False, "Disabled"]:
                return True
            if not enemy_options.get("revenge_limit_rando", False) in [
                False,
                "Vanilla",
            ]:
                return True

            return False

        if _should_run_khbr():
            return _run_khbr(
                self.extra_data.platform, enemy_options, mod.mod_yml, out_zip
            )
        else:
            return None, {}

    def generate_spoiler_html(
            self,
            enemy_spoilers_json,
            battle_level_spoiler,
            journal_hints_spoiler: dict[str, str]
    ) -> str:
        settings = self.settings
        randomizer = self.randomizer
        platform = self.extra_data.platform

        if platform == "Both":
            platform = "PC/PCSX2"

        exp_multipliers_json = {
            "Summon": {
                "multiplier": settings.summon_exp_multiplier,
                "values": list(accumulate(settings.summon_exp())),
            },
            "Valor": {
                "multiplier": settings.valor_exp_multiplier,
                "values": list(accumulate(settings.valor_exp())),
            },
            "Wisdom": {
                "multiplier": settings.wisdom_exp_multiplier,
                "values": list(accumulate(settings.wisdom_exp())),
            },
            "Limit": {
                "multiplier": settings.limit_exp_multiplier,
                "values": list(accumulate(settings.limit_exp())),
            },
            "Master": {
                "multiplier": settings.master_exp_multiplier,
                "values": list(accumulate(settings.master_exp())),
            },
            "Final": {
                "multiplier": settings.final_exp_multiplier,
                "values": list(accumulate(settings.final_exp())),
            },
        }
        sora_items_json = item_spoiler_dictionary(
            item_assignments=randomizer.assignments,
            starting_inventory_ids=randomizer.starting_item_ids,
            shop_items=randomizer.shop_items,
            weights=randomizer.location_weights,
            location_spheres=self.location_spheres,
        )
        donald_items_json = item_spoiler_dictionary(randomizer.donald_assignments)
        goofy_items_json = item_spoiler_dictionary(randomizer.goofy_assignments)
        synthesis_recipe_json = synth_recipe_dictionary(
            randomizer.assignments, randomizer.synthesis_recipes
        )
        settings_spoiler_json = self.settings.settings_spoiler_json
        weapon_stats_spoiler = weapon_stats_dictionary(
            sora_assignments=randomizer.assignments,
            donald_assignments=randomizer.donald_assignments,
            goofy_assignments=randomizer.goofy_assignments,
            weapons=randomizer.weapon_stats,
        )
        objectives_json = objectives_dictionary(randomizer.objectives)

        with open(resource_path("static/spoilerlog.html")) as spoiler_site:
            return (
                spoiler_site.read()
                .replace("{SEED_NAME_STRING}", settings.random_seed)
                .replace("{SEED_STRING}", settings.seed_string)
                .replace("{PLATFORM_GENERATED}", platform)
                .replace("{LEVEL_STATS_JSON}", json.dumps(levelStatsDictionary(randomizer.level_stats)))
                .replace("{FORM_EXP_JSON}", json.dumps(exp_multipliers_json))
                .replace("{DEPTH_VALUES_JSON}", json.dumps(randomizer.location_weights.weights))
                .replace("{SORA_ITEM_JSON}", json.dumps(sora_items_json, indent=4, cls=ItemEncoder))
                .replace("{DONALD_ITEM_JSON}", json.dumps(donald_items_json, indent=4, cls=ItemEncoder))
                .replace("{GOOFY_ITEM_JSON}", json.dumps(goofy_items_json, indent=4, cls=ItemEncoder))
                .replace("{BOSS_ENEMY_JSON}", json.dumps(enemy_spoilers_json))
                .replace("{BATTLE_LEVEL_JSON}", json.dumps(battle_level_spoiler))
                .replace("{SYNTHESIS_RECIPE_JSON}", json.dumps(synthesis_recipe_json))
                .replace("{WEAPON_STATS_JSON}", json.dumps(weapon_stats_spoiler))
                .replace("{JOURNAL_HINTS_JSON}", json.dumps(journal_hints_spoiler))
                .replace("{OBJECTIVES_JSON}", json.dumps(objectives_json))
                .replace("{SETTINGS_JSON}", json.dumps(settings_spoiler_json))
                .replace("PromiseCharm", "Promise Charm")
            )

    def generate_seed_hash_image(self, out_zip: ZipFile):
        hash_icons = self.settings.seedHashIcons
        image_data = hashimage.generate_seed_hash_image(hash_icons, use_bitmap=False)
        out_zip.writestr("preview.png", image_data)
        out_zip.writestr("misc/randoseed-hash-icons.csv", ",".join(hash_icons))

    def create_battle_level_rando_assets(
        self, mod: SeedModBuilder
    ) -> dict[locationType, list[int]]:
        settings = self.settings
        btlv_option_name = settings.battle_level_rando

        btlv = BtlvViewer()
        btlv.use_setting(
            btlv_option_name,
            battle_level_offset=settings.battle_level_offset,
            battle_level_range=settings.battle_level_range,
            battle_level_random_min_max=settings.battle_level_random_min_max,
        )
        if (
            (btlv_option_name == BattleLevelOption.NORMAL.name)
            or (
                btlv_option_name == BattleLevelOption.OFFSET.name
                and settings.battle_level_offset == 0
            )
            or (
                btlv_option_name == BattleLevelOption.RANDOM_WITHIN_RANGE.name
                and settings.battle_level_range == 0
            )
        ):
            return btlv.get_spoiler()

        modified_battle_level_binary = btlv.write_modifications()
        mod.write_battle_level_assets(modified_battle_level_binary)

        return btlv.get_spoiler()

    def add_cmd_list_modifications(self, mod: SeedModBuilder):
        settings = self.settings
        if settings.roxas_abilities_enabled and not settings.disable_final_form:
            cmd_yml = "better_stt/CmdList.yml"
        elif settings.roxas_abilities_enabled and settings.disable_final_form:
            cmd_yml = "better_stt/CmdListWDisableFinal.yml"
        elif not settings.roxas_abilities_enabled and settings.disable_final_form:
            cmd_yml = "disable_final_form.yml"
        else:  # Roxas abilities not enabled and final form not disabled - nothing to do
            return
        mod.write_cmd_list_modifications(modified_cmd_list_yml=cmd_yml)

    def create_puzzle_assets(self, mod: SeedModBuilder):
        if locationType.Puzzle not in self.settings.disabledLocations:
            assigned_puzzles = _assignment_subset_from_type(
                self.randomizer.assignments, [locationType.Puzzle]
            )
            with open(resource_path("static/puzzle.bin"), "rb") as puzzleBar:
                modified_puzzle_binary = bytearray(puzzleBar.read())
                for puzz in assigned_puzzles:
                    byte0 = 20 + puzz.location.LocationId * 16
                    byte1 = 20 + puzz.location.LocationId * 16 + 1
                    item = puzz.item.Id

                    item_byte_0, item_byte_1 = number_to_bytes(item)
                    modified_puzzle_binary[byte0] = item_byte_0
                    modified_puzzle_binary[byte1] = item_byte_1
                mod.write_puzzle_assets(modified_puzzle_binary)

    def create_drop_rate_assets(self, mod: SeedModBuilder):
        settings = self.settings
        global_jackpot = settings.global_jackpot
        global_lucky_lucky = settings.global_lucky
        fast_urns = settings.fast_urns
        rich_enemies = settings.rich_enemies
        near_unlimited_mp = settings.unlimited_mp

        if (
            global_jackpot > 0
            or global_lucky_lucky > 0
            or fast_urns
            or rich_enemies
            or near_unlimited_mp
        ):
            all_drops = {}
            testing = []
            with open(resource_path("static/drops.bin"), "rb") as drops_bar:
                drops_binary = bytearray(drops_bar.read())
                for i in range(0, 184):
                    start_index = 8 + 24 * i
                    rate = DropRates(
                        start_index, drops_binary[start_index : start_index + 24]
                    )
                    all_drops[rate.id] = rate

                    if rate.id not in id_to_enemy_name:
                        testing.append(rate.id)

            spawnable_enemy_ids = [
                1,
                2,
                3,
                4,
                5,
                6,
                7,
                8,
                9,
                10,
                11,
                12,
                13,
                15,
                17,
                18,
                22,
                23,
                24,
                25,
                26,
                27,
                28,
                29,
                30,
                32,
                33,
                34,
                35,
                36,
                37,
                38,
                39,
                40,
                41,
                42,
                43,
                44,
                45,
                46,
                47,
                48,
                49,
                50,
                51,
                52,
                63,
                64,
                65,
                66,
                69,
                73,
                87,
                88,
                89,
                96,
                112,
                171,
                172,
                173,
                174,
                175,
                176,
                177,
                178,
                179,
                180,
                181,
                182,
            ]
            urn_ids = [70, 71]
            stt_enemies = [119, 120, 121, 130, 145]
            struggles = [122, 131, 132]

            modded_ids = []

            if rich_enemies:
                for drop in all_drops.values():
                    if drop.id in spawnable_enemy_ids:
                        modded_ids.append(drop.id)
                        drop.medium_munny = max(drop.medium_munny, 2)
                        drop.small_munny = max(drop.small_munny, 2)
            if near_unlimited_mp:
                for drop in all_drops.values():
                    if drop.id in spawnable_enemy_ids:
                        modded_ids.append(drop.id)
                        drop.big_mp = max(drop.big_mp, 5)
                        drop.small_mp = max(drop.small_mp, 5)

            if global_lucky_lucky > 0:
                for drop in all_drops.values():
                    if drop.item1 != 0:
                        modded_ids.append(drop.id)
                        drop.item1_chance = min(
                            drop.item1_chance
                            + (drop.item1_chance // 2) * global_lucky_lucky,
                            100,
                        )
                    if drop.item2 != 0:
                        modded_ids.append(drop.id)
                        drop.item2_chance = min(
                            drop.item2_chance
                            + (drop.item2_chance // 2) * global_lucky_lucky,
                            100,
                        )
                    if drop.item3 != 0:
                        modded_ids.append(drop.id)
                        drop.item3_chance = min(
                            drop.item3_chance
                            + (drop.item3_chance // 2) * global_lucky_lucky,
                            100,
                        )
            if global_jackpot > 0:
                for drop in all_drops.values():
                    modded_ids.append(drop.id)
                    drop.small_hp = min(
                        drop.small_hp + (drop.small_hp // 2) * global_jackpot, 64
                    )
                    drop.big_hp = min(
                        drop.big_hp + (drop.big_hp // 2) * global_jackpot, 64
                    )
                    drop.big_munny = min(
                        drop.big_munny + (drop.big_munny // 2) * global_jackpot, 64
                    )
                    drop.medium_munny = min(
                        drop.medium_munny + (drop.medium_munny // 2) * global_jackpot,
                        64,
                    )
                    drop.small_munny = min(
                        drop.small_munny + (drop.small_munny // 2) * global_jackpot, 64
                    )
                    drop.small_mp = min(
                        drop.small_mp + (drop.small_mp // 2) * global_jackpot, 64
                    )
                    drop.big_mp = min(
                        drop.big_mp + (drop.big_mp // 2) * global_jackpot, 64
                    )
                    drop.small_drive = min(
                        drop.small_drive + (drop.small_drive // 2) * global_jackpot, 64
                    )
                    drop.big_drive = min(
                        drop.big_drive + (drop.big_drive // 2) * global_jackpot, 64
                    )

            if fast_urns:
                for u in urn_ids:
                    modded_ids.append(u)
                    all_drops[u].big_hp = 64

            for drop in all_drops.values():
                if drop.id in modded_ids:
                    mod.prize_table.add_prize(
                        identifier=drop.id,
                        small_hp_orbs=drop.small_hp,
                        big_hp_orbs=drop.big_hp,
                        big_money_orbs=drop.big_munny,
                        medium_money_orbs=drop.medium_munny,
                        small_money_orbs=drop.small_munny,
                        small_mp_orbs=drop.small_mp,
                        big_mp_orbs=drop.big_mp,
                        small_drive_orbs=drop.small_drive,
                        big_drive_orbs=drop.big_drive,
                        item_1=drop.item1,
                        item_1_percentage=drop.item1_chance,
                        item_2=drop.item2,
                        item_2_percentage=drop.item2_chance,
                        item_3=drop.item3,
                        item_3_percentage=drop.item3_chance,
                    )

    def create_objective_rando_assets(self, mod: SeedModBuilder, num_objectives_needed: int, objective_list: list[KH2Objective]):
        mod.add_objective_randomization_mods(num_objectives_needed, objective_list)
        mod.items.add_item(
            item_id=363,
            item_type="Recipe",
            flag_0=0,
            flag_1=0,
            rank="C",
            stat_entry=0,
            name=47883,
            description=47884,
            shop_buy=num_objectives_needed,
            shop_sell=0,
            command=0,
            slot=189,
            picture=226,
            icon_1=0,
            icon_2=2,
        )
    def create_emblem_rando_assets(self, mod: SeedModBuilder, num_emblems_needed: int):
        mod.add_emblem_randomization_mods(num_emblems_needed)
        mod.items.add_item(
            item_id=363,
            item_type="Recipe",
            flag_0=0,
            flag_1=0,
            rank="C",
            stat_entry=0,
            name=47883,
            description=47884,
            shop_buy=num_emblems_needed,
            shop_sell=0,
            command=0,
            slot=189,
            picture=226,
            icon_1=0,
            icon_2=2,
        )

    def create_shop_rando_assets(self, mod: SeedModBuilder):
        shop_items = self.randomizer.shop_items
        if len(shop_items) > 0:
            items_for_shop = []
            keyblade_item_ids = [
                i.Id for i in shop_items if i.ItemType == itemType.KEYBLADE
            ]
            report_item_ids = [
                i.Id for i in shop_items if i.ItemType == itemType.REPORT
            ]
            story_unlock_ids = [
                i.Id for i in shop_items if i.ItemType == itemType.STORYUNLOCK
            ]
            consumable_ids = [i.Id for i in shop_items if i.ItemType == itemType.ITEM]
            used_ids = (
                keyblade_item_ids + report_item_ids + story_unlock_ids + consumable_ids
            )
            remaining_items = [i.Id for i in shop_items if i.Id not in used_ids]

            if len(keyblade_item_ids) > 0:
                for keyblade_id in keyblade_item_ids:
                    items_for_shop.append((keyblade_id, 400))

            if len(report_item_ids) > 0:
                for report_id in report_item_ids:
                    items_for_shop.append((report_id, 75 * (report_id - 225)))
                for report_number in range(13):
                    short_name = f"Ansem Report {report_number + 1}"
                    mod.messages.add_message(
                        message_id=46778 - 32768 + report_number * 2, en=short_name
                    )

            if len(story_unlock_ids) > 0:
                for story_unlock_id in story_unlock_ids:
                    items_for_shop.append((story_unlock_id, 500))

            if len(consumable_ids) > 0:
                consumable_price_map = {
                    4: 400,
                    7: 600,
                    274: 400,
                    275: 600,
                    276: 250,
                    277: 250,
                    278: 250,
                    279: 250,
                }
                for consumable_id in consumable_ids:
                    if consumable_id in consumable_price_map:
                        items_for_shop.append(
                            (consumable_id, consumable_price_map[consumable_id])
                        )
                    else:
                        items_for_shop.append((consumable_id, 700))

            if len(remaining_items) > 0:
                raise GeneratorException(
                    f"Trying to put unexpected items in the shop: {len(remaining_items)}"
                )
                # price_map = {itemRarity.COMMON:100,itemRarity.UNCOMMON:300, itemRarity.RARE:500, itemRarity.MYTHIC:1000}
                # for i in remaining_items:
                #     items_for_shop.append((i.Id,price_map[i.Rarity]))

            with open(resource_path("static/full_items.json"), "r") as item_json:
                all_item_jsons = json.loads(item_json.read())
                for item_id, price in items_for_shop:
                    item_json = None
                    for y in all_item_jsons["Items"]:
                        if y["Id"] == item_id:
                            item_json = y
                            break
                    item_json["ShopBuy"] = price
                    mod.items.add_item(
                        item_id=item_json["Id"],
                        item_type=item_json["Type"],
                        flag_0=item_json["Flag0"],
                        flag_1=item_json["Flag1"],
                        rank=item_json["Rank"],
                        stat_entry=item_json["StatEntry"],
                        name=item_json["Name"],
                        description=item_json["Description"],
                        shop_buy=item_json["ShopBuy"],
                        shop_sell=item_json["ShopSell"],
                        command=item_json["Command"],
                        slot=item_json["Slot"],
                        picture=item_json["Picture"],
                        icon_1=item_json["Icon1"],
                        icon_2=item_json["Icon2"],
                    )

            with open(resource_path("static/shop.bin"), "rb") as shop_bar:
                modified_shop_binary = bytearray(shop_bar.read())

                byte0, byte1 = number_to_bytes(80 + len(items_for_shop))
                modified_shop_binary[10] = byte0
                modified_shop_binary[11] = byte1

                byte0, byte1 = number_to_bytes(len(items_for_shop))

                # inventory 752
                modified_shop_binary[754] = byte0
                modified_shop_binary[755] = byte1
                byte0, byte1 = number_to_bytes(984)
                modified_shop_binary[756] = byte0
                modified_shop_binary[757] = byte1

                valid_start = bytes_to_number(
                    modified_shop_binary[12], modified_shop_binary[13]
                )
                for asset in range(len(items_for_shop)):
                    product_index = 984 + 2 * asset
                    valid_item_index = valid_start + (60 + asset) * 2
                    byte0, byte1 = number_to_bytes(items_for_shop[asset][0])
                    modified_shop_binary[product_index] = byte0
                    modified_shop_binary[product_index + 1] = byte1
                    modified_shop_binary[valid_item_index] = byte0
                    modified_shop_binary[valid_item_index + 1] = byte1

                mod.write_shop_assets(modified_shop_binary)

                # ## code below prints out the shop information in relevant format

                # print(f"file type: {bytes_to_number(binaryContent[4],binaryContent[5])}")
                # shop_list_count = bytes_to_number(binaryContent[6],binaryContent[7])
                # print(f"shop list count: {shop_list_count}")
                # inventory_list_count = bytes_to_number(binaryContent[8],binaryContent[9])
                # print(f"inventory entry count: {inventory_list_count}")
                # product_list_count = bytes_to_number(binaryContent[10],binaryContent[11])
                # print(f"product entry count: {product_list_count}")
                # valid_start = bytes_to_number(binaryContent[12],binaryContent[13])
                # print(f"valid items offset: {valid_start}")

                # shop_start = 16
                # print("Shop Entries")
                # for x in range(shop_list_count):
                #     shop_index = shop_start+x*24
                #     print(f"---- Shop ID: {bytes_to_number(binaryContent[shop_index+18])}")
                #     print(f"---- Inventory Amount: {bytes_to_number(binaryContent[shop_index+16],binaryContent[shop_index+17])}")
                #     print(f"---- Inventory Offset: {bytes_to_number(binaryContent[shop_index+20],binaryContent[shop_index+21])}")
                #     print("----------")

                # inventory_start = shop_start+shop_list_count*24
                # print("Inventory Entries")
                # for x in range(inventory_list_count):
                #     inventory_index = inventory_start+x*8
                #     print(f"---- Inventory Address: {inventory_index}")
                #     print(f"---- Unlock event: {bytes_to_number(binaryContent[inventory_index],binaryContent[inventory_index+1])}")
                #     print(f"---- Product Amount: {bytes_to_number(binaryContent[inventory_index+2],binaryContent[inventory_index+3])}")
                #     print(f"---- Product Offset: {bytes_to_number(binaryContent[inventory_index+4],binaryContent[inventory_index+5])}")
                #     print("----------")

                # product_start = inventory_start+inventory_list_count*8
                # print("Product Entries")
                # for x in range(product_list_count):
                #     product_index = product_start+x*2
                #     print(f"---- Address {product_index}")
                #     print(f"---- Product (Item Id):  {bytes_to_number(binaryContent[product_index],binaryContent[product_index+1])}")

                # print("Valid Items")
                # for x in range(63):
                #     valid_index = valid_start+x*2
                #     item_id = bytes_to_number(binaryContent[valid_index],binaryContent[valid_index+1])
                #     if item_id!=0:
                #         print(f"---- Valid Item (Item Id):  {item_id}")

    def create_synth_assets(self, mod: SeedModBuilder):
        if locationType.SYNTH in self.settings.disabledLocations:
            return

        randomizer = self.randomizer
        assigned_synth = _assignment_subset_from_type(
            randomizer.assignments, [locationType.SYNTH]
        )

        synth_items = []
        for assignment in assigned_synth:
            synth_items.append(
                SynthLocation(
                    assignment.location.LocationId,
                    assignment.item.Id,
                    [
                        r
                        for r in randomizer.synthesis_recipes
                        if r.location == assignment.location
                    ][0],
                )
            )

        with open(resource_path("static/synthesis.bin"), "rb") as recipes_bar:
            modified_recipes_binary = bytearray(recipes_bar.read())
            for synth_loc in synth_items:
                starting_byte = synth_loc.get_starting_location()
                data = synth_loc.get_bytes()

                for i, item in enumerate(data):
                    modified_recipes_binary[starting_byte + i] = 0xFF & item

        with open(resource_path("static/synthesis_reqs.bin"), "rb") as requirements_bar:
            modified_requirements_binary = bytearray(requirements_bar.read())

            # uncomment to see some data about the synth lists
            # index = 16
            # while index + 12 < len(modified_requirements_binary):
            #     print(SynthList(index, modified_requirements_binary[index:]))
            #     index += 12

            # 3/6 free dev 8,9 bytes from offset index
            free_dev1 = number_to_bytes(3)
            free_dev2 = number_to_bytes(6)
            modified_requirements_binary[36] = free_dev1[0]
            modified_requirements_binary[37] = free_dev1[1]
            modified_requirements_binary[72] = free_dev2[0]
            modified_requirements_binary[73] = free_dev2[1]

            # 1,3 ori+ version,
            # free_dev1 = number_to_bytes(1)
            # free_dev2 = number_to_bytes(3)
            # ori_plus_id = number_to_bytes(12) # not sure on this
            # binaryContent[33] = ori_plus_id[0]
            # binaryContent[35] = 0
            # binaryContent[36] = free_dev1[0]
            # binaryContent[37] = free_dev1[1]
            # binaryContent[69] = ori_plus_id[0]
            # binaryContent[71] = 0
            # binaryContent[72] = free_dev2[0]
            # binaryContent[73] = free_dev2[1]

            # uncomment to make all existing synth buyable conditions need 7 of that material
            # new_required_items = number_to_bytes(7)
            # start_index=376
            # for i in range(0,24):
            #     binaryContent[start_index+i*12+8] = new_required_items[0]
            #     binaryContent[start_index+i*12+9] = new_required_items[1]

        mod.write_synth_assets(modified_recipes_binary, modified_requirements_binary)

    def create_chest_visual_assets(self, mod: SeedModBuilder):
        if not self.settings.chests_match_item:
            return

        chest_item_assignments = _assignment_subset(
            self.randomizer.assignments, [locationCategory.CHEST]
        )
        chest_item_assignments = [
            trsr
            for trsr in chest_item_assignments
            if locationType.Puzzle not in trsr.location.LocationTypes
            and locationType.Critical not in trsr.location.LocationTypes
            and locationType.SYNTH not in trsr.location.LocationTypes
        ]

        chests_by_location_id = ChestList.chests_by_location_id()
        chest_visual_assignments: list[ChestVisualAssignment] = []
        for chest_item_assignment in chest_item_assignments:
            location = chest_item_assignment.location
            location_types = location.LocationTypes

            item = chest_item_assignment.item
            item_type = item.ItemType

            # Just in case because D/G starting items are labeled as free and chests
            if locationType.Free in location_types and location.LocationId in range(
                1, 3
            ):
                # print('party stating item?')
                continue
            if (
                location.name() == stt.CheckLocation.StationOfSerenityPotion
                and item_type
                in [
                    itemType.PROOF_OF_CONNECTION,
                    itemType.PROOF_OF_PEACE,
                    itemType.PROOF_OF_NONEXISTENCE,
                    itemType.PROMISE_CHARM,
                ]
            ):
                # if this chest in STT Station of Serenity is a proof, it can't be big since it's a tutorial chest
                continue

            # use location id to get chest index and name
            chest = chests_by_location_id[location.LocationId]
            chest_visual_id = ChestList.chest_visual_id(location_types, item_type)
            chest_visual_assignments.append(
                ChestVisualAssignment(
                    chest.ChestIndex, chest.spawn_file_path(), chest_visual_id
                )
            )

        mod.write_chest_visuals_assets(chest_visual_assignments)

    def assign_starting_items(self, mod: SeedModBuilder):
        def pad_items(item_list):
            while len(item_list) < 32:
                item_list.append(0)

        settings = self.settings
        randomizer = self.randomizer
        master_item_list = (
            Items.getItemList()
            + Items.getActionAbilityList()
            + Items.getSupportAbilityList()
            + Items.getKeybladeAbilityList()
            + Items.getLevelAbilityList()
            + [Items.getTT1Jailbreak()]
            + [Items.objectiveReportItem()]
            + [Items.getPromiseCharm()]
        )
        growth_abilities = [
            i.Id
            for i in master_item_list
            if (
                i.ItemType == itemType.GROWTH_ABILITY
            )
        ]
        sora_abilities = [
            i.Id
            for i in master_item_list
            if (
                i.ItemType == itemType.GROWTH_ABILITY
                or i.ItemType == itemType.ACTION_ABILITY
                or i.ItemType == itemType.SUPPORT_ABILITY
            )
        ]

        donald_starting_items = (
            [1 + 0x8000, 3 + 0x8000]
            + [
                l.item.Id
                for l in _assignment_subset_from_type(
                    randomizer.donald_assignments, [locationType.Free]
                )
            ]
        )
        mod.player_params.add_player(
            character_id=2,  # Donald Starting Items
            identifier=0,
            hp=20,
            mp=100,
            ap=settings.donald_ap - 5,
            armor_slot_max=1,
            accessory_slot_max=2,
            item_slot_max=2,
            items=donald_starting_items,
            padding=[0] * 52,
        )

        goofy_starting_items = (
            [
                1 + 0x8000,
                1 + 0x8000,
                1 + 0x8000,
            ]
            + [
                l.item.Id
                for l in _assignment_subset_from_type(
                    randomizer.goofy_assignments, [locationType.Free]
                )
            ]
        )
        mod.player_params.add_player(
            character_id=3,  # Goofy Starting Items
            identifier=0,
            hp=20,
            mp=100,
            ap=settings.goofy_ap - 4,
            armor_slot_max=2,
            accessory_slot_max=1,
            item_slot_max=3,
            items=goofy_starting_items,
            padding=[0] * 52,
        )

        ability_equip = 0x8000 if settings.auto_equip_abilities else 0
        sora_bdscript_abilities = [
            i
            for i in randomizer.starting_item_ids
            if i in sora_abilities and i not in growth_abilities
        ]
        sora_bdscript_items = [
            i
            for i in randomizer.starting_item_ids
            if i not in sora_abilities
        ]

        mod.add_sora_bdscript(sora_bdscript_items,sora_bdscript_abilities,settings.auto_equip_abilities)

        sora_starting_items = [
            l.item.Id
            for l in _assignment_subset_from_type(
                randomizer.assignments, [locationType.Critical]
            )
        ] + [
            i + ability_equip
            for i in randomizer.starting_item_ids
            if i in sora_abilities and i in growth_abilities
        ]
        pad_items(sora_starting_items)
        mod.player_params.add_player(
            character_id=1,  # Sora Starting Items (Crit)
            identifier=7,  # crit difficulty
            hp=20,
            mp=100,
            ap=settings.sora_ap,
            armor_slot_max=1,
            accessory_slot_max=1,
            item_slot_max=3,
            items=sora_starting_items[:7],
            padding=[0] * 52,
        )

        lion_sora_items = [32930, 32930, 32931, 32931, 33288, 33289, 33290, 33294]
        pad_items(lion_sora_items)
        mod.player_params.add_player(
            character_id=135,  # Lion Dash on Lion Sora
            identifier=0,
            hp=0,
            mp=0,
            ap=0,
            armor_slot_max=0,
            accessory_slot_max=0,
            item_slot_max=0,
            items=lion_sora_items,
            padding=[0] * 52,
        )

        mod.player_params.add_player(
            character_id=1,  # Sora Starting Items (Non Crit)
            identifier=0,
            hp=20,
            mp=100,
            ap=settings.sora_ap,
            armor_slot_max=1,
            accessory_slot_max=1,
            item_slot_max=3,
            items=sora_starting_items[7:],
            padding=[0] * 52,
        )

    def assign_weapon_stats(self, mod: SeedModBuilder):
        randomizer = self.randomizer
        weapons = []
        weapons.extend(
            _assignment_subset(randomizer.assignments, [locationCategory.WEAPONSLOT])
        )
        weapons.extend(
            _assignment_subset(
                randomizer.donald_assignments, [locationCategory.WEAPONSLOT]
            )
        )
        weapons.extend(
            _assignment_subset(
                randomizer.goofy_assignments, [locationCategory.WEAPONSLOT]
            )
        )

        for weapon in weapons:
            weapon_stats = next(
                stat
                for stat in randomizer.weapon_stats
                if stat.location == weapon.location
            )
            mod.items.add_stats(
                location_id=weapon.location.LocationId,
                attack=weapon_stats.strength,
                magic=weapon_stats.magic,
                defense=0,
                ability=weapon.item.Id,
                ability_points=0,
                unknown_08=100,
                fire_resistance=100,
                ice_resistance=100,
                lightning_resistance=100,
                dark_resistance=100,
                unknown_0d=100,
                general_resistance=100,
                unknown=0,
            )

    def disable_antiform(self, mod: SeedModBuilder):
        mod.form_levels.turn_off_anti()

    def assign_form_levels(self, mod: SeedModBuilder):
        form_dict = {
            0: "Summon",
            1: "Valor",
            2: "Wisdom",
            3: "Limit",
            4: "Master",
            5: "Final",
        }
        form_category_list = [
            locationCategory.SUMMONLEVEL,
            locationCategory.VALORLEVEL,
            locationCategory.WISDOMLEVEL,
            locationCategory.LIMITLEVEL,
            locationCategory.MASTERLEVEL,
            locationCategory.FINALLEVEL,
        ]
        for index, levelType in enumerate(form_category_list):
            levels = _assignment_subset(self.randomizer.assignments, [levelType])
            form_name = form_dict[index]
            for level_number in range(1, 8):
                for lvl in levels:
                    if lvl.location.LocationId == level_number:
                        form_exp = next(
                            exp
                            for exp in self.randomizer.form_level_exp
                            if exp == lvl.location
                        )
                        mod.form_levels.add_form_level(
                            form_name=form_name,
                            form_id=index,
                            form_level=lvl.location.LocationId,
                            ability=lvl.item.Id
                            if index != 0
                            else 0,  # making summon junk items zero
                            experience=form_exp.experience,
                            growth_ability_level=0,
                        )

    def assign_bonuses(self, mod: SeedModBuilder):
        self._assign_sora_bonuses(mod, self.randomizer.assignments)
        self._assign_party_bonuses(mod, 2, "Donald", self.randomizer.donald_assignments)
        self._assign_party_bonuses(mod, 3, "Goofy", self.randomizer.goofy_assignments)

    @staticmethod
    def _assign_party_bonuses(
        mod: SeedModBuilder,
        character_id: int,
        character_name: str,
        assigned_items: list[ItemAssignment],
    ):
        for bon in _assignment_subset(
            assigned_items, locationCategory.bonus_categories()
        ):
            item1 = bon.item.Id
            item2 = 0
            if bon.item2 is not None:
                item2 = bon.item2.Id
            mod.bonuses.add_bonus(
                reward_id=bon.location.LocationId,
                character_id=character_id,
                character_name=character_name,
                hp_increase=0,
                mp_increase=0,
                drive_gauge_increase=0,
                item_slot_upgrade=0,
                accessory_slot_upgrade=0,
                armor_slot_upgrade=0,
                bonus_item_1=item1,
                bonus_item_2=item2,
                padding=0,
            )

    @staticmethod
    def _assign_sora_bonuses(mod: SeedModBuilder, assigned_items: list[ItemAssignment]):
        for bon in _assignment_subset(
            assigned_items, locationCategory.bonus_categories()
        ):
            character_id = 1  # Sora id
            character_name = "Sora"
            if locationType.STT in bon.location.LocationTypes:
                character_id = 14  # Roxas id
                character_name = "Roxas"

            # determine if assigned item is a stat bonus, and if so, use the bonuses native stat update
            hp_increase = 0
            mp_increase = 0
            drive_increase = 0
            item_slot_increase = 0
            accessory_slot_increase = 0
            armor_slot_increase = 0
            bonus_item_1 = 0
            bonus_item_2 = 0

            assigned_item_1 = bon.item.item
            if assigned_item_1 == bonus.MaxHpUp:
                hp_increase += 5
            elif assigned_item_1 == bonus.MaxMpUp:
                mp_increase += 10
            elif assigned_item_1 == bonus.DriveGaugeUp:
                drive_increase += 1
            elif assigned_item_1 == bonus.ArmorSlotUp:
                armor_slot_increase += 1
            elif assigned_item_1 == bonus.AccessorySlotUp:
                accessory_slot_increase += 1
            elif assigned_item_1 == bonus.ItemSlotUp:
                item_slot_increase += 1
            else:
                bonus_item_1 = assigned_item_1.id

            if bon.item2 is not None:
                assigned_item_2 = bon.item2.item
                if assigned_item_2 == bonus.MaxHpUp:
                    hp_increase += 5
                elif assigned_item_2 == bonus.MaxMpUp:
                    mp_increase += 10
                elif assigned_item_2 == bonus.DriveGaugeUp:
                    drive_increase += 1
                elif assigned_item_2 == bonus.ArmorSlotUp:
                    armor_slot_increase += 1
                elif assigned_item_2 == bonus.AccessorySlotUp:
                    accessory_slot_increase += 1
                elif assigned_item_2 == bonus.ItemSlotUp:
                    item_slot_increase += 1
                else:
                    bonus_item_2 = assigned_item_2.id

            mod.bonuses.add_bonus(
                reward_id=bon.location.LocationId,
                character_id=character_id,
                character_name=character_name,
                hp_increase=hp_increase,
                mp_increase=mp_increase,
                drive_gauge_increase=drive_increase,
                item_slot_upgrade=item_slot_increase,
                accessory_slot_upgrade=accessory_slot_increase,
                armor_slot_upgrade=armor_slot_increase,
                bonus_item_1=bonus_item_1,
                bonus_item_2=bonus_item_2,
                padding=0,
            )

    def assign_levels(self, mod: SeedModBuilder):
        settings = self.settings
        levels = _assignment_subset(
            self.randomizer.assignments, [locationCategory.LEVEL]
        )
        level_checks = settings.max_level_checks

        # get the triple of items for each level
        items_for_sword_level = {}
        offsets = DreamWeaponOffsets()
        for sword_level in range(1, 100):
            sword_item = 0
            shield_item = 0
            staff_item = 0
            shield_level = (
                offsets.get_item_lookup_for_shield(level_checks, sword_level)
                if settings.split_levels
                else sword_level
            )
            staff_level = (
                offsets.get_item_lookup_for_staff(level_checks, sword_level)
                if settings.split_levels
                else sword_level
            )
            for lvup in levels:
                if lvup.location.LocationId == sword_level:
                    sword_item = lvup.item.Id
                if lvup.location.LocationId == shield_level:
                    shield_item = lvup.item.Id
                if lvup.location.LocationId == staff_level:
                    staff_item = lvup.item.Id
            items_for_sword_level[sword_level] = (sword_item, shield_item, staff_item)

        for lvup in levels:
            level_stats = next(
                lv for lv in self.randomizer.level_stats if lv.location == lvup.location
            )
            level = lvup.location.LocationId
            item_id = items_for_sword_level[level]
            mod.level_ups.add_sora_level(
                level=level,
                experience=level_stats.experience,
                strength=level_stats.strength,
                magic=level_stats.magic,
                defense=level_stats.defense,
                ap=level_stats.ap,
                sword_ability=item_id[0],
                shield_ability=item_id[1],
                staff_ability=item_id[2],
                padding=0,
            )
            
        companion_names = ["Donald","Goofy","PingMulan","Beast","Auron","Sparrow","Aladdin","Jack","Simba","Tron","Riku"]
        companion_exp = settings.companion_exp()
        for c_name in companion_names:
            companion_levels = mod.level_ups.get_companion_levels(c_name)
            for level in companion_levels:
                mod.level_ups.add_companion_level(
                    level=level,
                    experience=companion_exp[level],
                    strength=companion_levels[level]["Strength"],
                    magic=companion_levels[level]["Magic"],
                    defense=companion_levels[level]["Defense"],
                    ap=companion_levels[level]["Ap"],
                    companion_name=c_name,
                    padding=0
                )

    def assign_treasures(self, mod: SeedModBuilder):
        treasures = _assignment_subset(
            self.randomizer.assignments,
            [locationCategory.POPUP, locationCategory.CHEST],
        )
        treasures = [
            trsr
            for trsr in treasures
            if locationType.Puzzle not in trsr.location.LocationTypes
            and locationType.Critical not in trsr.location.LocationTypes
            and locationType.SYNTH not in trsr.location.LocationTypes
        ]

        for trsr in treasures:
            mod.treasures.add_treasure(
                location_id=trsr.location.LocationId, item_id=trsr.item.Id
            )

    def _final_door_requirement_message(self) -> str:
        settings = self.settings

        message_lines: list[str] = []
        if settings.promiseCharm or misc.PromiseCharm.id in settings.starting_inventory_ids:
            message_lines.append("Entry requires")
            message_lines.append("Promise Charm and")
        else:
            message_lines.append("The Final Door")
            message_lines.append("opens with")

        if settings.objective_rando:
            message_lines.append(f"{settings.num_objectives_needed} completed objectives.")
        elif settings.emblems:
            message_lines.append(f"{settings.num_emblems_needed} Lucky Emblems.")
        else:
            message_lines.append("3 Proofs.")

        return "\n".join(message_lines)

    def check_if_companion_changed(
        self, settings: SeedSettings, keys: settingkey
    ):
        return settings.get(keys.COMPANION_DAMAGE_TOGGLE)

    def prepare_companion_damage_knockback(self, mod: SeedModBuilder):
        keys = settingkey
        atkp_organizer = mod._get_atkp_organizer()
        ui_settings = self.settings.ui_settings
        companions_changed = self.check_if_companion_changed(ui_settings, keys)
        knockback_organizer = KnockbackTypes.get_knockback_value

        if not companions_changed:
            return
        kill_boss = 0
        if ui_settings.get(keys.COMPANION_KILL_BOSS):
            kill_boss = "KillBoss"

        # Donald
        options_list = []
        options_list.append(
            knockback_organizer(
                ui_settings.get(keys.DONALD_MELEE_ATTACKS_KNOCKBACK_TYPE)
            )
        )
        options_list.append(
            knockback_organizer(ui_settings.get(keys.DONALD_FIRE_KNOCKBACK_TYPE))
        )
        options_list.append(
            knockback_organizer(
                ui_settings.get(keys.DONALD_BLIZZARD_KNOCKBACK_TYPE)
            )
        )
        options_list.append(
            knockback_organizer(ui_settings.get(keys.DONALD_THUNDER_KNOCKBACK_TYPE))
        )
        self.ready_companion_damage_knockback_atkp_entries(
            atkp_organizer, "Donald/PL_Donald", options_list, kill_boss
        )
        
        # Goofy
        options_list = []
        options_list.append(
            knockback_organizer(
                ui_settings.get(keys.GOOFY_MELEE_ATTACKS_KNOCKBACK_TYPE)
            )
        )
        options_list.append(
            knockback_organizer(ui_settings.get(keys.GOOFY_BASH_KNOCKBACK_TYPE))
        )
        options_list.append(
            knockback_organizer(ui_settings.get(keys.GOOFY_TURBO_KNOCKBACK_TYPE))
        )
        options_list.append(
            knockback_organizer(ui_settings.get(keys.GOOFY_TORNADO_KNOCKBACK_TYPE))
        )
        self.ready_companion_damage_knockback_atkp_entries(
            atkp_organizer, "Goofy/PL_Goofy", options_list, kill_boss
        )
        
        # Jack Skellington
        options_list = []
        options_list.append(
            knockback_organizer(
                ui_settings.get(keys.JACK_SKELLINGTON_MELEE_ATTACKS_KNOCKBACK_TYPE)
            )
        )
        options_list.append(
            knockback_organizer(ui_settings.get(keys.JACK_SKELLINGTON_BLAZING_FURY_KNOCKBACK_TYPE))
        )
        options_list.append(
            knockback_organizer(ui_settings.get(keys.JACK_SKELLINGTON_ICY_TERROR_KNOCKBACK_TYPE))
        )
        options_list.append(
            knockback_organizer(ui_settings.get(keys.JACK_SKELLINGTON_BOLTS_OF_SORROW_KNOCKBACK_TYPE))
        )
        self.ready_companion_damage_knockback_atkp_entries(
            atkp_organizer, "Jack Skellington", options_list, kill_boss
        )
        
        # Simba
        options_list = []
        options_list.append(
            knockback_organizer(
                ui_settings.get(keys.SIMBA_MELEE_ATTACKS_KNOCKBACK_TYPE)
            )
        )
        options_list.append(
            knockback_organizer(ui_settings.get(keys.SIMBA_FIERCE_CLAW_KNOCKBACK_TYPE))
        )
        options_list.append(
            knockback_organizer(ui_settings.get(keys.SIMBA_GROUNDSHAKER_KNOCKBACK_TYPE))
        )
        self.ready_companion_damage_knockback_atkp_entries(
            atkp_organizer, "Simba", options_list, kill_boss
        )

        # Aladdin
        options_list = []
        options_list.append(
            knockback_organizer(
                ui_settings.get(keys.ALADDIN_MELEE_ATTACKS_KNOCKBACK_TYPE)
            )
        )
        options_list.append(
            knockback_organizer(ui_settings.get(keys.ALADDIN_SLASH_FRENZY_KNOCKBACK_TYPE))
        )
        options_list.append(
            knockback_organizer(ui_settings.get(keys.ALADDIN_QUICKPLAY_KNOCKBACK_TYPE))
        )
        self.ready_companion_damage_knockback_atkp_entries(
            atkp_organizer, "Aladdin", options_list, kill_boss
        )
        
        # Ping/Mulan
        options_list = []
        options_list.append(
            knockback_organizer(
                ui_settings.get(keys.PING_MULAN_MELEE_ATTACKS_KNOCKBACK_TYPE)
            )
        )
        options_list.append(
            knockback_organizer(ui_settings.get(keys.MULAN_MUSHU_FIRE_KNOCKBACK_TYPE))
        )
        options_list.append(
            knockback_organizer(ui_settings.get(keys.MULAN_FLAMETONGUE_KNOCKBACK_TYPE))
        )
        self.ready_companion_damage_knockback_atkp_entries(
            atkp_organizer, "Ping/Mulan", options_list, kill_boss
        )
        
        # Beast
        options_list = []
        options_list.append(
            knockback_organizer(
                ui_settings.get(keys.BEAST_MELEE_ATTACKS_KNOCKBACK_TYPE)
            )
        )
        options_list.append(
            knockback_organizer(ui_settings.get(keys.BEAST_SHOUT_KNOCKBACK_TYPE))
        )
        options_list.append(
            knockback_organizer(ui_settings.get(keys.BEAST_RUSH_KNOCKBACK_TYPE))
        )
        self.ready_companion_damage_knockback_atkp_entries(
            atkp_organizer, "Beast", options_list, kill_boss
        )
        
        # Tron
        options_list = []
        options_list.append(
            knockback_organizer(
                ui_settings.get(keys.TRON_MELEE_ATTACKS_KNOCKBACK_TYPE)
            )
        )
        options_list.append(
            knockback_organizer(ui_settings.get(keys.TRON_SCOUTING_DISK_KNOCKBACK_TYPE))
        )
        options_list.append(
            knockback_organizer(ui_settings.get(keys.TRON_PULSING_THUNDER_KNOCKBACK_TYPE))
        )
        self.ready_companion_damage_knockback_atkp_entries(
            atkp_organizer, "Tron", options_list, kill_boss
        )
        
        # Jack Sparrow
        options_list = []
        options_list.append(
            knockback_organizer(
                ui_settings.get(keys.JACK_SPARROW_MELEE_ATTACKS_KNOCKBACK_TYPE)
            )
        )
        options_list.append(
            knockback_organizer(ui_settings.get(keys.JACK_SPARROW_NO_MERCY_KNOCKBACK_TYPE))
        )
        options_list.append(
            knockback_organizer(ui_settings.get(keys.JACK_SPARROW_RAIN_STORM_KNOCKBACK_TYPE))
        )
        options_list.append(
            knockback_organizer(ui_settings.get(keys.JACK_SPARROW_BONE_SMASH_KNOCKBACK_TYPE))
        )
        self.ready_companion_damage_knockback_atkp_entries(
            atkp_organizer, "Jack Sparrow", options_list, kill_boss
        )
        
        # Riku
        options_list = []
        options_list.append(
            knockback_organizer(
                ui_settings.get(keys.RIKU_MELEE_ATTACKS_KNOCKBACK_TYPE)
            )
        )
        options_list.append(
            knockback_organizer(ui_settings.get(keys.RIKU_DARK_AURA_KNOCKBACK_TYPE))
        )
        options_list.append(
            knockback_organizer(ui_settings.get(keys.RIKU_DARK_SHIELD_KNOCKBACK_TYPE))
        )
        self.ready_companion_damage_knockback_atkp_entries(
            atkp_organizer, "Riku", options_list, kill_boss
        )
        
        # Auron
        options_list = []
        options_list.append(
            knockback_organizer(
                ui_settings.get(keys.AURON_MELEE_ATTACKS_KNOCKBACK_TYPE)
            )
        )
        options_list.append(
            knockback_organizer(ui_settings.get(keys.AURON_DIVIDER_KNOCKBACK_TYPE))
        )
        self.ready_companion_damage_knockback_atkp_entries(
            atkp_organizer, "Auron", options_list, kill_boss
        )

    def ready_companion_damage_knockback_atkp_entries(
        self,
        atkp_organizer: AttackEntriesOrganizer,
        companion_name,
        melee_ability_knockback_types_list,
        kill_boss,
    ):
        companion_melee_ids = self._get_companion_ids_for_damage_knockback_options(
            companion_name, "Melee"
        )
        companion_melee_objects = []
        for melee_entry in companion_melee_ids:
            if len(melee_entry) == 3:
                companion_melee_objects.append(
                    atkp_organizer.get_attack_using_ids_plus_switch(melee_entry[0], melee_entry[1], melee_entry[2])    
                )
            companion_melee_objects.append(
                atkp_organizer.get_attack_using_ids(melee_entry[0], melee_entry[1])
            )
        for melee_object in companion_melee_objects:
            melee_object: ATKPObject
            melee_object.EnemyReaction = melee_ability_knockback_types_list[0]
            melee_object.Flags = kill_boss

        companion_ability_ids = self._get_companion_ids_for_damage_knockback_options(
            companion_name, "Ability"
        )
        companion_ability_objects = []
        for i in range(len(companion_ability_ids)):
            # The third entry is for when we need the switch field to distinguish different entries for the same attack (like goofy tornado)
            for ability_entry in companion_ability_ids[i]:
                if len(ability_entry) == 3:
                    entry_object = atkp_organizer.get_attack_using_ids_plus_switch(
                        ability_entry[0], ability_entry[1], ability_entry[2]
                    )
                    entry_object: ATKPObject
                    entry_object.EnemyReaction = melee_ability_knockback_types_list[i + 1]
                    entry_object.Flags = kill_boss
                    companion_ability_objects.append(entry_object)
                    continue
                
                entry_object = atkp_organizer.get_attack_using_ids(
                    ability_entry[0], ability_entry[1]
                )
                entry_object: ATKPObject
                entry_object.EnemyReaction = melee_ability_knockback_types_list[i+1]
                entry_object.Flags = kill_boss
                companion_ability_objects.append(entry_object)

        for atkp_object in companion_melee_objects:
            atkp_organizer.convert_atkp_object_to_dict_and_add_to_data(atkp_object)
        for atkp_object in companion_ability_objects:
            atkp_organizer.convert_atkp_object_to_dict_and_add_to_data(atkp_object)

    def _get_companion_ids_for_damage_knockback_options(
        self, companion_name, id_group
    ) -> dict[{str, list[int]}]:
        # SubId first, then Id
        if companion_name == "Donald/PL_Donald":
            if id_group == "Melee":
                return [[0, 151], [0, 152], [0, 153], [0, 154], [0, 155], [0, 153], [0, 154]]
            return [
                [[0, 1163], [2, 1163], [0, 1716], [2, 1716]],
                [[0, 1164], [0, 1717]],
                [[0, 1165, 1], [0, 1165, 2], [0, 1718, 1], [0, 1718, 2]],
            ]
        elif companion_name == "Goofy/PL_Goofy":
            if id_group == "Melee":
                return [[0, 146], [1, 146], [0, 156], [0, 157], [0, 158], [0, 159], [0, 1942], [1, 1942], [0, 85], [0, 86], [0, 87], [0, 95]]
            return [
                [[0, 1161]],
                [[0, 1162], [0, 1253]],
                [[0, 1160, 1], [0, 1258, 1]],
            ]
        elif companion_name == "Jack Skellington":
            if id_group == "Melee":
                return [[0, 224], [0, 895], [0, 225], [0, 563], [0,564], [0, 565], [0, 273], [1, 273], [0, 274]]
            return [
                [[0, 1176]],
                [[0, 1177, 1], [0, 1177, 2]],
                [[0, 1178]],
            ]
        elif companion_name == "Simba":
            if id_group == "Melee":
                return [[0, 861], [0, 941], [0, 862], [0, 863], [0, 864], [0, 864]]
            return [
                [[0, 1180]],
                [[0, 1179], [1, 1179], [4, 1179]],
            ]
        elif companion_name == "Aladdin":
            if id_group == "Melee":
                return [[0, 214], [0, 215], [0, 216], [0, 217], [0, 218]]
            return [
                [[0, 1174], [1, 1174]],
                [[0, 1175], [0, 1193], [1, 1193]], # Quickplay spans over 2 parent IDs
            ]
        elif companion_name == "Ping/Mulan":
            if id_group == "Melee":
                return [[0, 365], [0, 366], [0, 367], [0, 368], [0, 369], [1, 369], [0, 371], [0, 384], [0, 387], [0, 385], [0, 372], [0, 373]]
            return [
                [[0, 1169], [3, 1169]],
                [[0, 622], [1, 622]],
            ]
        elif companion_name == "Beast":
            if id_group == "Melee":
                return [[0, 219], [0, 220], [0, 223], [0, 221], [0, 222]]
            return [
                [[0, 1166]],
                [[0, 570], [0, 575]], # Rush spans over 2 parent IDs
            ]
        elif companion_name == "Tron":
            if id_group == "Melee":
                return [[0, 361], [0, 1527], [0, 1528], [0, 1529]]
            return [
                [[0, 627]],
                [[0, 1181]],
            ]
        elif companion_name == "Jack Sparrow":
            if id_group == "Melee":
                return [[0, 494], [0, 495], [0, 646], [0, 647], [0, 648], [0, 648], [0, 515], [0, 649]]
            return [
                [[0, 1171]],
                [[0, 1172]],
                [[0, 1173]],
            ]
        elif companion_name == "Riku":
            if id_group == "Melee":
                return [[0, 1494], [0, 1495], [0, 1496], [0, 1497], [0, 1498], [0, 1499], [0, 1500], [0, 1501], [0, 1502], [0, 1503], [0, 1504], [0, 1505], [0, 1506], [0, 1507], [0, 1508], [0, 1509], [0, 1510], [0, 1511], [0, 1512], [0, 1513], [0, 1514], [0, 1515], [0, 1516]]
            return [
                [[0, 1531]],
                [[0, 1532]]
            ]
        elif companion_name == "Auron":
            if id_group == "Melee":
                return [[0, 193], [0, 194], [0, 195], [0, 196]]
            return [
                [[0, 1272], [1, 1272], [0, 1576]], # Divider spans over 2 parent ids
            ]


class CosmeticsOnlyZip:

    def __init__(self, ui_settings: SeedSettings):
        self.settings = ui_settings

    def create_zip(self) -> io.BytesIO:
        data = io.BytesIO()
        with ZipFile(data, "w", ZIP_DEFLATED) as out_zip:
            mod = ModYml(
                "Randomized Cosmetics",
                description="Generated by the KH2 Randomizer Seed Generator.",
            )

            _add_cosmetics(out_zip=out_zip, mod_yml=mod, settings=self.settings)

            mod.write_to_zip_file(out_zip)

            out_zip.write(
                resource_path("static/icons/misc/Kingdom Hearts II.png"), "icon.png"
            )

        data.seek(0)
        return data


class BossEnemyOnlyZip:
    def __init__(self, seed_name: str, ui_settings: SeedSettings, platform: str):
        self.settings = ui_settings
        self.enemy_options = makeKHBRSettings(seed_name, self.settings)
        self.platform = platform

    def create_zip(self) -> io.BytesIO:
        def _should_run_khbr():
            if not self.enemy_options.get("boss", False) in [False, "Disabled"]:
                return True
            if not self.enemy_options.get("enemy", False) in [False, "Disabled"]:
                return True
            return False

        if not _should_run_khbr():
            raise GeneratorException(
                "Trying to generate boss/enemy only mod without enabling those settings."
            )

        data = io.BytesIO()
        with ZipFile(data, "w", ZIP_DEFLATED) as out_zip:
            mod = ModYml(
                "Randomized Bosses/Enemies",
                description="Generated by the KH2 Randomizer Seed Generator.",
            )

            enemy_spoilers, _ = _run_khbr(
                self.platform, self.enemy_options, mod, out_zip
            )

            mod.write_to_zip_file(out_zip)

            out_zip.write(
                resource_path("static/icons/misc/Kingdom Hearts II.png"), "icon.png"
            )
            out_zip.writestr("enemyspoilers.txt", enemy_spoilers)

            data.seek(0)
            return data



import base64
import copy
import json
import random
import textwrap
from itertools import permutations, chain
from typing import Optional, Any
from zipfile import ZipFile

from Class.exceptions import HintException
from Class.itemClass import KH2Item
from Class.newLocationClass import KH2Location
from List.ItemList import Items
from List.configDict import itemType, locationType, locationCategory, HintType
from List.inventory import ability, form, magic, misc, storyunlock, summon, proof
from List.location import weaponslot
from Module import version
from Module.Hints.HintOutput import HintOutput
from Module.Hints.HintUtils import (
    CommonTrackerInfo,
    HintUtils,
    JsmarteeHintData,
    PathHintData,
    PointHintData,
    WorldItems,
)
from Module.RandomizerSettings import RandomizerSettings
from Module.newRandomize import Randomizer
from Module.seedmod import SeedModBuilder


HintData = dict[str, Any]


class Hints:

    @staticmethod
    def generator_journal_hints(
        location_item_data: list[tuple[KH2Location, KH2Item]],
        settings: RandomizerSettings,
        hint_data: HintData,
    ):
        journal_data: dict[int, str] = {}
        for report_num in range(1, 14):
            journal_data[report_num] = ""
        # find any settings relating to journal hints
        independent_hints = settings.journal_hints != "Off"
        independent_hint_specific = settings.journal_hints == "exact"

        # if independent hints, need to make "useful" query on the randomized locations and items
        if independent_hints:
            Hints.independent_journal_hints(location_item_data, journal_data, independent_hint_specific)

        # add journal text to hintsText structure
        if "Reports" not in hint_data:
            hint_data["Reports"] = {}
            for report_num in range(1, 14):
                hint_data["Reports"][report_num] = {}
        for report_num in range(1, 14):
            if report_num not in hint_data["Reports"]:
                hint_data["Reports"][report_num] = {}
            hint_data["Reports"][report_num]["JournalText"] = journal_data[report_num]

    @staticmethod
    def get_journal_hint(
            location: KH2Location,
            item_data: KH2Item,
            location_item_data: list[tuple[KH2Location, KH2Item]],
            independent_hint_specific: bool,
    ) -> Optional[str]:
        """
        If applicable, returns hint text to be used in the journal for the given item at the given location. Returns
        None if the location is not hintable or the item is not one that should receive a journal hint.
        """

        location_id = location.LocationId
        if location_id == weaponslot.LocationId.KingdomKey:
            # Kinda useless to hint Kingdom Key
            return None

        items_to_find = [
            ability.Scan,
            ability.AerialRecovery,
            ability.Guard,
            ability.ComboMaster,
            ability.FinishingPlus,
            ability.ExperienceBoost,
            ability.BerserkCharge,
            ability.HorizontalSlash,
            ability.Slapshot,
            ability.FinishingLeap,
            ability.FlashStep,
            ability.SlideDash,
            ability.GuardBreak,
            ability.Explosion,
            ability.AerialSpiral,
            ability.AerialDive,
            ability.MagnetBurst,
            ability.AutoValor,
            ability.AutoWisdom,
            ability.AutoLimitForm,
            ability.AutoMaster,
            ability.AutoFinal,
            ability.TrinityLimit,
            ability.NegativeCombo,
        ]
        if item_data.item not in items_to_find:
            return None

        hintable_worlds = [
            locationType.Level,
            locationType.LoD,
            locationType.BC,
            locationType.HB,
            locationType.TT,
            locationType.TWTNW,
            locationType.SP,
            locationType.Atlantica,
            locationType.PR,
            locationType.OC,
            locationType.Agrabah,
            locationType.HT,
            locationType.PL,
            locationType.DC,
            locationType.HUNDREDAW,
            locationType.STT,
            locationType.FormLevel,
        ]

        hinted_item_name = item_data.Name

        # if ability is on keyblade, check location of keyblade
        if locationCategory.WEAPONSLOT in location.LocationCategory:
            special_key_to_location = {
                weaponslot.LocationId.KingdomKeyD: "Valor Form",
                weaponslot.LocationId.AlphaWeapon: "Master Form",
                weaponslot.LocationId.OmegaWeapon: "Final Form"
            }
            if location_id in special_key_to_location:
                # Form keys we don't give additional information to avoid hinting the forms themselves
                return f"{hinted_item_name} is on {special_key_to_location[location_id]}'s Keyblade."

            # find the location of the keyblade
            keyblade_item = Items.weaponslot_id_to_keyblade_item(location_id)
            keyblade_location = next(
                (key_loc for key_loc, key_item in location_item_data if key_item.Name == keyblade_item.name), None
            )
            if keyblade_location is None:
                # Not in a world we're allowed to hint
                return None
            hintable_world = next(
                (world for world in keyblade_location.LocationTypes if world in hintable_worlds), None
            )
            if hintable_world is None:
                # Not in a world we're allowed to hint
                return None

            world_text = HintUtils.location_hint_user_friendly_text(hintable_world)
            if independent_hint_specific:
                key_location_text = f"{world_text} - {keyblade_location.Description}"
                return f"{hinted_item_name} is at {key_location_text} (on {keyblade_item.name})."
            else:
                return f"{hinted_item_name} is in {world_text} (on {keyblade_item.name})."
        else:
            hintable_world = next((world for world in location.LocationTypes if world in hintable_worlds), None)
            if hintable_world is None:
                # Not in a world we're allowed to hint
                return None

            world_text = HintUtils.location_hint_user_friendly_text(hintable_world)
            if independent_hint_specific:
                return f"{hinted_item_name} is at {world_text} - {location.Description}."
            else:
                return f"{hinted_item_name} is in {world_text}."

    @staticmethod
    def independent_journal_hints(
        location_item_data: list[tuple[KH2Location, KH2Item]],
        journal_data: dict[int, str],
        independent_hint_specific: bool,
    ):
        all_possible_independent_hints = []
        for location, item_data in location_item_data:
            hint = Hints.get_journal_hint(location, item_data, location_item_data, independent_hint_specific)
            if hint is not None:
                all_possible_independent_hints.append(hint)

        # at this point, we should have all our hint text, just need to assign to each report
        report_index = 0
        while len(all_possible_independent_hints):
            journal_data[report_index + 1] += (all_possible_independent_hints[-1] + "\n\n")
            report_index = (report_index + 1) % 13
            all_possible_independent_hints.pop()

    @staticmethod
    def generate_hints_v2(randomizer: Randomizer, settings: RandomizerSettings) -> HintData:
        # this list is meant to disallow worlds from being hinted, since they will never have hintable items
        exclude_list = HintUtils.update_disabled_worlds_on_tracker(settings)
        location_item_tuples = HintUtils.convert_item_assignment_to_tuple(
            randomizer.assignments, randomizer.shop_items
        )

        common_tracker_data = CommonTrackerInfo(settings)
        world_items = WorldItems(location_item_tuples, common_tracker_data)
        hintable_worlds = [
            world for world in HintUtils.hintable_worlds() if world not in exclude_list
        ]
        hint_data = common_tracker_data.to_dict()
        hint_data["level_data"] = world_items.level_checks
        hint_data["world"] = world_items.world_to_item_ids()
        if settings.emblems:
            hint_data["emblems"] = {"max_emblems_available":settings.max_emblems_available,"num_emblems_needed":settings.num_emblems_needed}

        if settings.objective_rando:
            objective_locations = [o.Location for o in randomizer.objectives]
            # find these locations in location_item_tuples
            objective_tuples = [{"category":l_i[0].LocationCategory, "location_id":l_i[0].LocationId} for l_i in location_item_tuples if l_i[0].Description in objective_locations]
            # remove duplicates
            objective_tuples = [dict(t) for t in {tuple(d.items()) for d in objective_tuples}]
            
            hint_data["objective_list"] = [o.Name for o in randomizer.objectives]
            hint_data["objective_locations"] = objective_tuples
            hint_data["num_objectives_needed"] = settings.num_objectives_needed
        if settings.hintsType == HintType.SHANANAS:
            hint_data["Reports"] = copy.deepcopy(world_items.report_information)
            if common_tracker_data.progression_settings is not None:
                world_list = list(hint_data["world"].keys())
                random.shuffle(world_list)
                hint_data["world_order"] = world_list
        elif settings.hintsType == HintType.JSMARTEE:
            jsmartee_data = []
            for world in hintable_worlds:
                jsmartee_data.append(JsmarteeHintData(world_items, world))
            if common_tracker_data.progression_settings is not None:
                hint_data["Reports"] = HintUtils.jsmartee_progression_hints(
                    world_items, jsmartee_data
                )
            else:
                hint_data["Reports"] = HintUtils.jsmartee_hint_report_assignment(
                    settings, world_items, jsmartee_data
                )
        elif settings.hintsType == HintType.POINTS:
            if common_tracker_data.progression_settings is not None:
                world_list = list(hint_data["world"].keys())
                random.shuffle(world_list)
                hint_data["world_order"] = world_list
            point_data = []
            for world in hintable_worlds:
                point_data.append(
                    PointHintData(settings, common_tracker_data, world_items, world)
                )
            hint_data["Reports"] = HintUtils.point_hint_report_assignment(
                settings, world_items, point_data
            )
        elif settings.hintsType == HintType.SPOILER:
            hint_data["reveal_data"] = world_items.revealed_item_ids_to_names()
            hint_data["aux_data"] = world_items.item_ids_to_names()
            hint_data["Reports"] = HintUtils.spoiler_hint_assignment(
                settings, common_tracker_data, world_items, hintable_worlds
            )
        elif settings.hintsType == HintType.PATH:
            path_data = []
            for world in HintUtils.hintable_worlds():
                path_data.append(PathHintData(world_items, world))
            hint_data["Reports"] = HintUtils.path_hint_assignment(
                world_items,
                path_data,
                common_tracker_data,
                hintable_worlds,
            )
        elif settings.hintsType == HintType.DISABLED:
            # don't need to do anything extra
            pass
        else:
            raise HintException("Unable to generate hints for nonexistent hint system")

        hint_data["startingInventory"] = randomizer.starting_item_ids

        Hints.generator_journal_hints(location_item_tuples, settings, hint_data)
        return hint_data

    @staticmethod
    def write_hints(hint_data: HintData, out_zip: ZipFile):
        # too many things in unhinted now to not send the tracker a file
        # if hint_data["hintsType"] != HintType.DISABLED:
        json_bytes = json.dumps(hint_data).encode("utf-8")
        if version.debug_mode():
            out_zip.writestr("HintFile_DebugHints.json", json_bytes)
        out_zip.writestr(
            "HintFile.Hints", base64.b64encode(json_bytes).decode("utf-8")
        )

    @staticmethod
    def write_hint_text(hint_data: HintData, mod: SeedModBuilder, journal_hints_spoiler: dict[str, str]):
        def convert_string_to_unicode(string: str, newlines: int = 0):
            return (
                    "".join(r"\u{:04X}".format(ord(character)) for character in textwrap.fill(string, width=30))
                    + "NEWLINE" * newlines
            )

        def hinted_world_text(input_hinted_world: Any) -> str:
            if isinstance(input_hinted_world, locationType):
                return HintUtils.location_hint_user_friendly_text(input_hinted_world)
            else:
                return str(input_hinted_world)

        for report_number in range(0, 13):
            report_data = hint_data["Reports"][report_number + 1]

            spoiler_text = ""
            encoded_text = ""

            hint_text = []
            if "ProgressionSettings" in hint_data:
                if hint_data["hintsType"] == HintType.POINTS:
                    hinted_world = report_data["World"]
                    hint_text.append(f"{hinted_world_text(hinted_world)} has {report_data['itemName']}.")
            else:
                if hint_data["hintsType"] == HintType.JSMARTEE:
                    hinted_world = report_data["World"]
                    check_count = report_data["Count"]
                    if check_count == 1:
                        hint_text.append(f"{hinted_world_text(hinted_world)} has {check_count} important check.")
                    else:
                        hint_text.append(f"{hinted_world_text(hinted_world)} has {check_count} important checks.")

                if hint_data["hintsType"] == HintType.PATH:
                    hint_text.append(report_data["Text"])

                if hint_data["hintsType"] == HintType.POINTS:
                    hinted_world = report_data["World"]
                    hint_text.append(f"{hinted_world_text(hinted_world)} has {report_data['itemName']}.")

                if hint_data["hintsType"] == HintType.SPOILER:
                    hinted_world = report_data["World"]
                    hint_text.append(f"{hinted_world_text(hinted_world)} has:")
                    for c in hint_data["world"][hinted_world]:
                        if c in hint_data["reveal_data"]:
                            c_txt = hint_data["reveal_data"][c]
                            hint_text.append(c_txt)

            for hint_text_instance in hint_text:
                spoiler_text += hint_text_instance + "\n"
                encoded_text += convert_string_to_unicode(hint_text_instance, newlines=1)
            spoiler_text += "\n"
            encoded_text += convert_string_to_unicode("", newlines=1)

            if "JournalText" in report_data:
                for line in report_data["JournalText"].splitlines():
                    stripped = line.strip()
                    if len(stripped) > 0:
                        spoiler_text += stripped + "\n\n"
                        encoded_text += convert_string_to_unicode(stripped, newlines=2)

            if "Location" in report_data:
                report_location = report_data["Location"]
                if isinstance(report_location, locationType):
                    report_location_text = HintUtils.location_hint_user_friendly_text(report_location)
                elif report_location == "":
                    report_location_text = "Sora's pocket"
                else:
                    report_location_text = report_location
                location_text = f"This report was found in {report_location_text}."
                spoiler_text += location_text + "\n\n"
                encoded_text += convert_string_to_unicode(location_text, newlines=2)

            if report_number < 10:
                report_message_id = 14052 + report_number * 2
            else:  # for some god awful reason, report 11-13 aren't contiguous with 1-10
                report_message_id = 14255 + (report_number - 10) * 2

            mod.journal_txt.add_message(
                message_id=report_message_id,
                en=encoded_text
            )

            journal_hints_spoiler[f"Report {report_number + 1}"] = spoiler_text.strip()



import json
import os
import random
from pathlib import Path

from Class.openkhmod import Asset
from List import configDict
from Module import appconfig


class Itempic:

    def __init__(self, itempic_id: str, item_name: str, types: list[str], custom: bool = False):
        super().__init__()
        self.itempic_id = itempic_id
        self.item_name = item_name
        self.types = types
        self.custom = custom

    def filename(self) -> str:
        return f"remastered/itempic/item-{self.itempic_id}.imd/-0.dds"


class ItempicRandomizer:

    @staticmethod
    def itempic_rando_options() -> dict[str, str]:
        return {
            configDict.VANILLA: "Vanilla",
            configDict.RANDOMIZE_IN_GAME_ONLY: "Randomize (in-game only)",
            configDict.RANDOMIZE_CUSTOM_ONLY: "Randomize (custom only)",
            configDict.RANDOMIZE_ALL: "Randomize (in-game + custom)",
        }

    @staticmethod
    def bootstrap_itempic_file() -> Path:
        """Creates the itempic list file if needed."""
        itempic_list_file_path = Path("item-pictures.json")
        if not itempic_list_file_path.is_file():
            with open(itempic_list_file_path, mode="w", encoding="utf-8") as itempic_list_file:
                mapped = []
                for itempic in ItempicRandomizer.replaceable_itempics():
                    mapped.append({
                        "id": itempic.itempic_id,
                        "name": itempic.item_name,
                        "type": itempic.types,
                    })
                json.dump(mapped, itempic_list_file, indent=4)
        return itempic_list_file_path

    @staticmethod
    def randomize_itempics(setting: str) -> list[Asset]:
        if setting == configDict.VANILLA:
            return []

        # We could maybe give an option here, but not sure if it's necessary
        categorize = True

        replacements_by_category: dict[str, list[tuple[bool, str]]] = {}  # (internal, path)

        if setting == configDict.RANDOMIZE_IN_GAME_ONLY or setting == configDict.RANDOMIZE_ALL:
            for category, game_files in ItempicRandomizer._collect_game_itempic_files(categorize).items():
                if category not in replacements_by_category:
                    replacements_by_category[category] = []
                replacements_by_category[category].extend((True, file) for file in game_files)

        if setting == configDict.RANDOMIZE_CUSTOM_ONLY or setting == configDict.RANDOMIZE_ALL:
            for category, custom_files in ItempicRandomizer._collect_custom_itempic_files(categorize).items():
                if category not in replacements_by_category:
                    replacements_by_category[category] = []
                replacements_by_category[category].extend((False, str(path)) for path in custom_files)

        backup_files_by_categories: dict[str, list[tuple[bool, str]]] = {}
        for category, replacements in replacements_by_category.items():
            backup_files_by_categories[category] = replacements.copy()
            random.shuffle(replacements)

        assets: list[Asset] = []

        itempic_list_file_path = ItempicRandomizer.bootstrap_itempic_file()
        with open(itempic_list_file_path, encoding="utf-8") as itempic_list_file:
            itempic_metadata = json.load(itempic_list_file)
        random.shuffle(itempic_metadata)
        for info in itempic_metadata:
            itempic_id: str = info["id"]
            types: list[str] = [
                item_type.lower() for item_type in info["type"] if item_type.lower() in replacements_by_category
            ]
            if len(types) > 0:
                random.shuffle(types)

                for chosen_type in types:
                    itempics_for_chosen_type = replacements_by_category.get(chosen_type, [])

                    # Unlike the music rando we'll just always allow duplicate replacements here to simplify
                    if len(itempics_for_chosen_type) == 0:
                        refill_list = backup_files_by_categories.get(chosen_type, []).copy()
                        random.shuffle(refill_list)
                        replacements_by_category[chosen_type] = refill_list
                        itempics_for_chosen_type = refill_list

                    if len(itempics_for_chosen_type) > 0:
                        internal, chosen_itempic = itempics_for_chosen_type.pop()

                        asset = {
                            "name": f"remastered/itempic/item-{itempic_id}.imd/-0.dds",
                            "platform": "pc",
                            "method": "copy",
                            "source": [{"name": chosen_itempic}]
                        }
                        if internal:
                            asset["source"][0]["type"] = "internal"

                        assets.append(asset)

                        break

        return assets

    @staticmethod
    def replaceable_itempics() -> list[Itempic]:
        def consumable(itempic_id: str, item_name: str) -> Itempic:
            return Itempic(itempic_id=itempic_id, item_name=item_name, types=["Consumable", "Consumable", "Wild"])

        def keyblade(itempic_id: str, item_name: str, custom: bool = False) -> Itempic:
            return Itempic(itempic_id=itempic_id, item_name=item_name, types=["Keyblade"], custom=custom)

        def staff(itempic_id: str, item_name: str) -> Itempic:
            return Itempic(itempic_id=itempic_id, item_name=item_name, types=["Staff"])

        def shield(itempic_id: str, item_name: str) -> Itempic:
            return Itempic(itempic_id=itempic_id, item_name=item_name, types=["Shield"])

        def armor(itempic_id: str, item_name: str) -> Itempic:
            return Itempic(itempic_id=itempic_id, item_name=item_name, types=["Armor", "Armor", "Wild"])

        def accessory(itempic_id: str, item_name: str) -> Itempic:
            return Itempic(itempic_id=itempic_id, item_name=item_name, types=["Accessory", "Accessory", "Wild"])

        def synthesis(itempic_id: str, item_name: str) -> Itempic:
            return Itempic(itempic_id=itempic_id, item_name=item_name, types=["Synthesis", "Synthesis", "Wild"])

        result = [
            Itempic("000", "Map (Large)", types=["Map"]),

            consumable("001", "Potion"),
            consumable("002", "Hi-Potion"),
            consumable("003", "Mega Potion"),
            consumable("004", "Ether"),
            consumable("005", "Hi-Ether"),
            consumable("006", "Mega Ether"),
            consumable("007", "Elixir"),
            consumable("008", "Megalixir"),
            consumable("009", "Drive Recovery"),
            consumable("010", "High Drive Recovery"),

            Itempic("011", "Map (Small)", types=["Map"]),

            Itempic("012", "Action Ability", types=["ActionAbility"], custom=True),
            Itempic("013", "Growth Ability", types=["GrowthAbility"], custom=True),
            Itempic("014", "Support Ability", types=["SupportAbility"], custom=True),
            Itempic("015", "Dummy 23 (HP Up)", types=["StatIncrease", "StatIncrease", "Wild"], custom=True),
            Itempic("016", "Dummy 24 (MP Up)", types=["StatIncrease", "StatIncrease", "Wild"], custom=True),
            Itempic("017", "Dummy 25 (Drive Up)", types=["StatIncrease", "StatIncrease", "Wild"], custom=True),
            Itempic("018", "Dummy 26 (Armor Slot Up)", types=["StatIncrease", "StatIncrease", "Wild"], custom=True),
            Itempic("019", "Dummy 27 (Accessory Slot Up)", types=["StatIncrease", "StatIncrease", "Wild"], custom=True),
            Itempic("020", "Dummy 16 (Item Slot Up)", types=["StatIncrease", "StatIncrease", "Wild"], custom=True),

            Itempic("021", "Torn Pages", types=["TornPage"]),

            consumable("033", "Tent"),
            consumable("034", "Power Boost"),
            consumable("035", "Magic Boost"),
            consumable("036", "Defense Boost"),
            consumable("037", "AP Boost"),

            keyblade("043", "Kingdom Key"),
            keyblade("044", "Oathkeeper"),
            keyblade("045", "Oblivion"),
            keyblade("046", "Star Seeker"),
            keyblade("047", "Hidden Dragon"),
            keyblade("048", "Hero's Crest"),
            keyblade("049", "Monochrome"),
            keyblade("050", "Follow the Wind"),
            keyblade("051", "Circle of Life"),
            keyblade("052", "Photon Debugger"),
            keyblade("053", "Gull Wing"),
            keyblade("054", "Rumbling Rose"),
            keyblade("055", "Guardian Soul"),
            keyblade("056", "Wishing Lamp"),
            keyblade("057", "Decisive Pumpkin"),
            keyblade("058", "Sleeping Lion"),
            keyblade("059", "Sweet Memories"),
            keyblade("060", "Mysterious Abyss"),
            keyblade("061", "Fatal Crest"),
            keyblade("062", "Bond of Flame"),
            keyblade("063", "Fenrir"),
            keyblade("064", "Ultima Weapon"),
            keyblade("065", "Struggle Hammer"),
            keyblade("066", "Struggle Wand"),
            keyblade("067", "Struggle Sword"),

            staff("068", "Mage's Staff"),
            staff("069", "Hammer Staff"),
            staff("070", "Victory Bell"),
            staff("071", "Meteor Staff"),
            staff("072", "Comet Staff"),
            staff("073", "Lord's Broom"),
            staff("074", "Wisdom Wand"),
            staff("075", "Rising Dragon"),
            staff("076", "Nobody Lance"),
            staff("077", "Shaman's Relic"),
            staff("078", "Save the Queen"),

            shield("080", "Knight's Shield"),
            shield("081", "Adamant Shield"),
            shield("082", "Chain Gear"),
            shield("083", "Ogre Shield"),
            shield("084", "Falling Star"),
            shield("085", "Dreamcloud"),
            shield("086", "Knight Defender"),
            shield("087", "Genji Shield"),
            shield("088", "Akashic Record"),
            shield("089", "Nobody Guard"),
            shield("090", "Save the King"),

            Itempic("091", "Dummy 13 (Royal Summons)", types=["RoyalSummons"], custom=True),
            Itempic("092", "Scimitar", types=["Scimitar"]),
            Itempic("093", "Battlefields of War", types=["BattlefieldsOfWar"]),
            Itempic("094", "Sword of the Ancestor", types=["SwordOfTheAncestor"]),
            Itempic("095", "Identity Disk", types=["IdentityDisk"]),
            Itempic("096", "Skill and Crossbones", types=["SkillAndCrossbones"]),
            Itempic("097", "Way to the Dawn", types=["WayToTheDawn"]),
            Itempic("098", "Beast's Claw", types=["BeastsClaw"]),
            Itempic("099", "Bone Fist", types=["BoneFist"]),
            Itempic("100", "Proud Fang", types=["ProudFang"]),

            armor("101", "Elven Bandana"),
            armor("102", "Divine Bandana"),
            armor("103", "Power Band"),
            armor("104", "Buster Band"),
            armor("105", "Protect Belt"),
            armor("106", "Gaia Belt"),
            armor("107", "Cosmic Belt"),
            armor("113", "Fire Bangle"),
            armor("114", "Fira Bangle"),
            armor("115", "Firaga Bangle"),
            armor("116", "Firagun Bangle"),
            armor("118", "Blizzard Armlet"),
            armor("119", "Blizzara Armlet"),
            armor("120", "Blizzaga Armlet"),
            armor("121", "Blizzagun Armlet"),
            armor("123", "Thunder Trinket"),
            armor("124", "Thundara Trinket"),
            armor("125", "Thundaga Trinket"),
            armor("126", "Thundagun Trinket"),
            armor("128", "Shadow Anklet"),
            armor("129", "Dark Anklet"),
            armor("130", "Midnight Anklet"),
            armor("131", "Chaos Anklet"),
            armor("133", "Abas Chain"),
            armor("134", "Aegis Chain"),
            armor("135", "Acrisius"),
            armor("136", "Ribbon"),
            armor("137", "Champion Belt"),
            armor("138", "Petite Ribbon"),
            armor("139", "Acrisius+"),
            armor("140", "Cosmic Chain"),

            accessory("149", "Ability Ring"),
            accessory("150", "Engineer's Ring"),
            accessory("151", "Technician's Ring"),
            accessory("152", "Expert's Ring"),
            accessory("153", "Sardonyx Ring"),
            accessory("154", "Tourmaline Ring"),
            accessory("155", "Aquamarine Ring"),
            accessory("156", "Garnet Ring"),
            accessory("157", "Diamond Ring"),
            accessory("158", "Silver Ring"),
            accessory("159", "Gold Ring"),
            accessory("160", "Platinum Ring"),
            accessory("161", "Mythril Ring"),
            accessory("162", "Orichalcum Ring"),
            accessory("163", "Master's Ring"),
            accessory("164", "Moon Amulet"),
            accessory("165", "Star Charm"),
            accessory("166", "Skill Ring"),
            accessory("167", "Skillful Ring"),
            accessory("168", "Soldier Earring"),
            accessory("169", "Fencer Earring"),
            accessory("170", "Mage Earring"),
            accessory("171", "Slayer Earring"),
            accessory("172", "Cosmic Ring"),
            accessory("173", "Medal"),
            accessory("174", "Cosmic Arts"),
            accessory("177", "Lucky Ring"),
            accessory("179", "Draw Ring"),

            synthesis("181", "Dark Shard"),
            synthesis("182", "Dark Stone"),
            synthesis("183", "Dark Gem"),
            synthesis("184", "Dark Crystal"),
            synthesis("185", "Blazing Shard"),
            synthesis("186", "Blazing Stone"),
            synthesis("187", "Blazing Gem"),
            synthesis("188", "Blazing Crystal"),
            synthesis("189", "Frost Shard"),
            synthesis("190", "Frost Stone"),
            synthesis("191", "Frost Gem"),
            synthesis("192", "Frost Crystal"),
            synthesis("193", "Lightning Shard"),
            synthesis("194", "Lightning Stone"),
            synthesis("195", "Lightning Gem"),
            synthesis("196", "Lightning Crystal"),
            synthesis("197", "Power Shard"),
            synthesis("198", "Power Stone"),
            synthesis("199", "Power Gem"),
            synthesis("200", "Power Crystal"),
            synthesis("201", "Lucid Shard"),
            synthesis("202", "Lucid Stone"),
            synthesis("203", "Lucid Gem"),
            synthesis("204", "Lucid Crystal"),
            synthesis("205", "Dense Shard"),
            synthesis("206", "Dense Stone"),
            synthesis("207", "Dense Gem"),
            synthesis("208", "Dense Crystal"),
            synthesis("209", "Twilight Shard"),
            synthesis("210", "Twilight Stone"),
            synthesis("211", "Twilight Gem"),
            synthesis("212", "Twilight Crystal"),
            synthesis("213", "Mythril Shard"),
            synthesis("214", "Mythril Stone"),
            synthesis("215", "Mythril Gem"),
            synthesis("216", "Mythril Crystal"),
            synthesis("217", "Bright Shard"),
            synthesis("218", "Bright Stone"),
            synthesis("219", "Bright Gem"),
            synthesis("220", "Bright Crystal"),

            Itempic("221", "Lamp Charm", types=["LampCharm"]),
            Itempic("222", "Ukulele Charm", types=["UkuleleCharm"]),
            Itempic("223", "Feather Charm", types=["FeatherCharm"]),
            Itempic("224", "Baseball Charm", types=["BaseballCharm"]),
            Itempic("225", "Munny Pouch", types=["MunnyPouch"]),
            Itempic("226", "Lucky Emblem", types=["LuckyEmblem"]),
            Itempic("227", "Struggle Trophy", types=["OtherItem", "OtherItem", "Wild"]),
            Itempic("228", "Struggle Poster", types=["OtherItem", "OtherItem", "Wild"]),
            Itempic("229", "Poster", types=["OtherItem", "OtherItem", "Wild"]),
            Itempic("230", "???", types=["OtherItem", "OtherItem", "Wild"]),
            Itempic("231", "Namine's Sketches", types=["NaminesSketches"]),
            Itempic("232", "Membership Card", types=["MembershipCard"]),
            Itempic("233", "Olympus Stone", types=["OlympusStone"]),
            Itempic("234", "Auron's Statue", types=["OtherItem", "OtherItem", "Wild"]),
            Itempic("235", "Cursed Medallion", types=["OtherItem", "OtherItem", "Wild"]),
            Itempic("236", "Present", types=["OtherItem", "OtherItem", "Wild"]),
            Itempic("237", "Decoy Presents", types=["OtherItem", "OtherItem", "Wild"]),
            Itempic("238", "Ice Cream", types=["IceCream"]),
            Itempic("239", "Picture", types=["OtherItem", "OtherItem", "Wild"]),

            Itempic("240", "Dummy 15 (Unknown Disk)", types=["UnknownDisk"], custom=True),

            Itempic("241", "Secret Ansem Report", types=["AnsemReport"]),
            Itempic("241", "Cursed Medallion", types=["OtherItem", "OtherItem", "Wild"]),
            Itempic("242", "Promise Charm", types=["PromiseCharm"]),
            Itempic("243", "Synthesis Recipe", types=["Recipe"]),
            Itempic("244", "Hades Cup Trophy", types=["HadesCupTrophy"]),

            Itempic("266", "Valor Form", types=["ValorForm"]),
            Itempic("267", "Wisdom Form", types=["WisdomForm"]),
            Itempic("268", "Master Form", types=["MasterForm"]),
            Itempic("269", "Final Form", types=["FinalForm"], custom=True),
            Itempic("270", "Magic Element", types=["MagicElement"]),

            Itempic("271", "Anti Form", types=["AntiForm"], custom=True),

            synthesis("272", "Energy Shard"),
            synthesis("273", "Energy Stone"),
            synthesis("274", "Energy Gem"),
            synthesis("275", "Energy Crystal"),
            synthesis("276", "Serenity Shard"),
            synthesis("277", "Serenity Stone"),
            synthesis("278", "Serenity Gem"),
            synthesis("279", "Serenity Crystal"),
            synthesis("280", "Orichalcum"),
            synthesis("281", "Orichalcum+"),

            keyblade("296", "Pureblood", custom=True),
            keyblade("297", "Alpha Weapon", custom=True),
            keyblade("298", "Omega Weapon", custom=True),
            keyblade("299", "Kingdom Key D", custom=True),

            keyblade("300", "Two Become One"),
            keyblade("301", "Winner's Proof"),

            staff("302", "Save the Queen+"),
            staff("303", "Centurion"),
            staff("304", "Centurion+"),
            staff("305", "Plain Mushroom"),
            staff("306", "Plain Mushroom+"),
            staff("307", "Precious Mushroom"),
            staff("308", "Precious Mushroom+"),
            staff("309", "Premium Mushroom"),

            shield("310", "Save the King+"),
            shield("311", "Frozen Pride"),
            shield("312", "Frozen Pride+"),
            shield("313", "Joyous Mushroom"),
            shield("314", "Joyous Mushroom+"),
            shield("315", "Majestic Mushroom"),
            shield("316", "Majestic Mushroom+"),
            shield("317", "Ultimate Mushroom"),

            staff("318", "Shaman's Relic+"),
            shield("319", "Akashic Record+"),

            Itempic("320", "Limit Form", types=["LimitForm"]),
            Itempic("321", "Proof of Connection", types=["ProofOfConnection"]),
            Itempic("322", "Proof of Nonexistence", types=["ProofOfNonexistence"]),
            Itempic("323", "Proof of Peace", types=["ProofOfPeace"]),

            Itempic("324", "Fire Element", types=["FireElement"], custom=True),
            Itempic("325", "Blizzard Element", types=["BlizzardElement"], custom=True),
            Itempic("326", "Thunder Element", types=["ThunderElement"], custom=True),
            Itempic("327", "Cure Element", types=["CureElement"], custom=True),
            Itempic("328", "Magnet Element", types=["MagnetElement"], custom=True),
            Itempic("329", "Reflect Element", types=["ReflectElement"], custom=True),

            armor("330", "Shock Charm"),
            armor("331", "Shock Charm+"),
            armor("332", "Grand Ribbon"),

            accessory("340", "Shadow Archive"),
            accessory("341", "Shadow Archive+"),
            accessory("342", "Full Bloom"),
            accessory("344", "Executive's Ring"),

            synthesis("350", "Remembrance Shard"),
            synthesis("351", "Remembrance Stone"),
            synthesis("352", "Remembrance Gem"),
            synthesis("353", "Remembrance Crystal"),
            synthesis("354", "Tranquility Shard"),
            synthesis("355", "Tranquility Stone"),
            synthesis("356", "Tranquility Gem"),
            synthesis("357", "Tranquility Crystal"),
            synthesis("358", "Manifest Illusion"),
            synthesis("359", "Lost Illusion"),
        ]

        return result

    @staticmethod
    def _collect_game_itempic_files(categorize: bool) -> dict[str, list[str]]:
        """Returns game itempic files grouped by category."""
        result: dict[str, list[str]] = {}

        for itempic in ItempicRandomizer.replaceable_itempics():
            if itempic.custom:
                continue

            resolved_category = itempic.types[0].lower()
            if not categorize:
                resolved_category = "wild"
            if resolved_category not in result:
                result[resolved_category] = []
            result[resolved_category].append(itempic.filename())

        return result

    @staticmethod
    def _collect_custom_itempic_files(categorize: bool) -> dict[str, list[Path]]:
        """Returns custom itempic files grouped by category."""
        result: dict[str, list[Path]] = {}

        custom_visuals_path = appconfig.read_custom_visuals_path()
        if custom_visuals_path is not None:
            item_pictures_path = custom_visuals_path / "item-pictures"
            if item_pictures_path.is_dir():
                for child in [str(category_file).lower() for category_file in os.listdir(item_pictures_path)]:
                    child_path = item_pictures_path / child
                    if child_path.is_dir():
                        category_pics: list[Path] = []
                        for root, dirs, files in os.walk(child_path):
                            root_path = Path(root)
                            for file in files:
                                _, extension = os.path.splitext(file)
                                if extension.lower() == ".dds" or extension.lower() == ".png":
                                    file_path = root_path / file
                                    category_pics.append(file_path)
                        resolved_category = child
                        if not categorize:
                            resolved_category = "wild"
                        if resolved_category not in result:
                            result[resolved_category] = []
                        result[resolved_category] += category_pics

        return result



from List.configDict import itemType, itemRarity
from dataclasses import dataclass, field
from json import JSONEncoder

from List.inventory.item import InventoryItem


@dataclass(frozen=True)
class KH2Item:
    item: InventoryItem
    Rarity: itemRarity = itemRarity.COMMON
    Id: int = field(init=False)
    Name: str = field(init=False)
    ItemType: itemType = field(init=False)

    def __post_init__(self):
        # Working around the frozen-ness here to set these
        object.__setattr__(self, "Id", self.item.id)
        object.__setattr__(self, "Name", self.item.name)
        object.__setattr__(self, "ItemType", self.item.type)


class ItemEncoder(JSONEncoder):

    def default(self, o):
        # Keep the JSON the same as it was previously even though we changed the structure a bit
        if isinstance(o, KH2Item):
            return {
                "Id": o.Id,
                "Name": o.Name,
                "ItemType": o.ItemType,
                "Rarity": o.Rarity
            }
        else:
            return o.__dict__



import os
import random
from pathlib import Path

from Class.openkhmod import Asset
from List import configDict
from Module import appconfig

AGRABAH = "al0"
BEAST_CASTLE = "bb0"
PORT_ROYAL = "ca0"
DISNEY_CASTLE = "dc0"
LINGERING_WILL = "dc1"
DESTINY_ISLANDS = "di0"
TWTNW = "eh0"
ROXAS = "eh1"
AGRABAH_ALT = "es0"
HOLLOW_BASTION = "hb0"
GARDEN_OF_ASSEMBLAGE = "hb1"
ABSENT_SILHOUETTE = "hb2"
OLYMPUS_COLISEUM = "he0"
UNDERWORLD = "he1"
PRIDE_LANDS = "lk0"
ATLANTICA = "lm0"
LAND_OF_DRAGONS = "mu0"
HALLOWEEN_TOWN = "nm0"
CHRISTMAS_TOWN = "nm1"
HUNDRED_ACRE_WOOD = "po0"
SPACE_PARANOIDS = "tr0"
TWILIGHT_TOWN = "tt0"
STATION_OF_CALLING = "tt1"
MYSTERIOUS_TOWER = "tt2"
MANSION_BASEMENT = "tt3"
WHITE_ROOM = "tt4"
MANSION = "tt5"
BETWIXT_BETWEEN = "tt6"
TIMELESS_RIVER = "wi0"
WORLD_MAP = "wm0"
KINGDOM_HEARTS_1 = "zz0"


class CommandMenuRandomizer:

    def __init__(self, command_menu_choice: str):
        super().__init__()
        self.command_menu_choice = command_menu_choice

    @staticmethod
    def command_menu_options() -> dict[str, str]:
        return {
            configDict.VANILLA: "Vanilla",
            configDict.RANDOMIZE_ONE: "Randomize (one)",
            configDict.RANDOMIZE_ALL: "Randomize (all)",
            AGRABAH: "Agrabah",
            BEAST_CASTLE: "Beast's Castle",
            PORT_ROYAL: "Port Royal",
            DISNEY_CASTLE: "Disney Castle",
            LINGERING_WILL: "Lingering Will",
            # DESTINY_ISLANDS: "Destiny Island",  # no visible difference from KH1
            TWTNW: "The World That Never Was",
            ROXAS: "Roxas Fight",  # looks like KH1 but with classic icons
            # AGRABAH_ALT: "Agrabah (alt),  # for carpet minigames? need to check for differences.
            HOLLOW_BASTION: "Hollow Bastion",
            GARDEN_OF_ASSEMBLAGE: "Garden of Assemblage",
            ABSENT_SILHOUETTE: "Absent Silhouette",
            OLYMPUS_COLISEUM: "Olympus Coliseum",
            UNDERWORLD: "The Underworld",
            PRIDE_LANDS: "Pride Lands",
            ATLANTICA: "Atlantica (PCSX2 Only)",
            LAND_OF_DRAGONS: "Land of Dragons",
            HALLOWEEN_TOWN: "Halloween Town",
            CHRISTMAS_TOWN: "Christmas Town",
            HUNDRED_ACRE_WOOD: "100 Acre Wood",
            SPACE_PARANOIDS: "Space Paranoids",
            TWILIGHT_TOWN: "Twilight Town",
            STATION_OF_CALLING: "Station of Calling",  # looks like kh1 but with classic icons.
            MYSTERIOUS_TOWER: "Mysterious Tower",
            MANSION_BASEMENT: "Mansion Basement",
            WHITE_ROOM: "The White Room",
            MANSION: "Mansion",
            BETWIXT_BETWEEN: "Betwixt & Between",
            TIMELESS_RIVER: "Timeless River",
            # WORLD_MAP: "World Map",  # no visible difference from Betwixt & Between
            KINGDOM_HEARTS_1: "Kingdom Hearts 1"
        }

    def randomize_command_menus(self) -> list[Asset]:
        if self.command_menu_choice == configDict.VANILLA:
            return []
        assets = []
        assets.extend(self._randomize_ps2())
        assets.extend(self._randomize_pc())
        return assets

    def _randomize_ps2(self) -> list[Asset]:
        menu_replacements = self._compute_menu_replacements(pc=False)
        assets: list[Asset] = []
        for old_menu, new_menu in menu_replacements.items():
            assets.append({
                "platform": "ps2",
                "name": f"field2d/jp/{old_menu}command.2dd",
                "multi": [
                    {"name": f"field2d/us/{old_menu}command.2dd"}
                ],
                "method": "copy",
                "source": [
                    {
                        "name": f"field2d/jp/{new_menu}command.2dd",
                        "type": "internal"
                    }
                ]
            })
        return assets

    def _randomize_pc(self) -> list[Asset]:
        menu_replacements = self._compute_menu_replacements(pc=True)
        assets: list[Asset] = []
        for old_menu, new_menu in menu_replacements.items():
            if new_menu == ATLANTICA:
                # Unsupported on PC - just leave it vanilla
                continue

            for region in ["fr", "gr", "it", "sp", "us"]:
                assets.append({
                    "platform": "pc",
                    "name": f"field2d/{region}/{old_menu}command.2dd",
                    "method": "copy",
                    "source": [
                        {
                            "name": f"field2d/{region}/{new_menu}command.2dd",
                            "type": "internal"
                        }
                    ]
                })

                remastered_region = f"remastered/field2d/{region}"
                cap_region = region.upper()
                assets.append({
                    "platform": "pc",
                    "name": f"{remastered_region}/{old_menu}command.2dd/{cap_region}_{old_menu}command_2dd0.png",
                    "method": "copy",
                    "source": [
                        {
                            "name": f"{remastered_region}/{new_menu}command.2dd/{cap_region}_{new_menu}command_2dd0.png",
                            "type": "internal"
                        }
                    ]
                })
        return assets

    def _compute_menu_replacements(self, pc: bool) -> dict[str, str]:
        unsupported = [configDict.VANILLA, configDict.RANDOMIZE_ONE, configDict.RANDOMIZE_ALL]
        if pc:
            unsupported.append(ATLANTICA)
        supported_menus = [menu for menu in self.command_menu_options().keys() if menu not in unsupported]

        command_menu_choice = self.command_menu_choice

        menu_replacements: dict[str, str] = {}
        if command_menu_choice == configDict.RANDOMIZE_ALL:
            shuffled_menus = supported_menus.copy()
            random.shuffle(shuffled_menus)
            for index, old_menu in enumerate(supported_menus):
                menu_replacements[old_menu] = shuffled_menus[index]
        elif command_menu_choice == configDict.RANDOMIZE_ONE:
            new_menu = random.choice(supported_menus)
            for old_menu in supported_menus:
                menu_replacements[old_menu] = new_menu
        else:  # A specific command menu was chosen
            for old_menu in supported_menus:
                menu_replacements[old_menu] = command_menu_choice
        return menu_replacements


class RoomTransitionImageRandomizer:

    def __init__(self, transition_choice: str):
        super().__init__()
        self.transition_choice = transition_choice

    @staticmethod
    def room_transition_options() -> dict[str, str]:
        return {
            configDict.VANILLA: "Vanilla",
            configDict.RANDOMIZE_IN_GAME_ONLY: "Randomize (in-game only)",
            configDict.RANDOMIZE_CUSTOM_ONLY: "Randomize (custom only)",
            configDict.RANDOMIZE_ALL: "Randomize (in-game + custom)",
        }

    def randomize_room_transitions(self) -> list[Asset]:
        if self.transition_choice == configDict.VANILLA:
            return []
        assets = []
        # assets.extend(self._randomize_ps2())
        assets.extend(self._randomize_pc())
        return assets

    def _randomize_pc(self) -> list[Asset]:
        transition_replacements = self._compute_transition_replacements()
        assets: list[Asset] = []
        for old_image, new_image in transition_replacements.items():
            if new_image.startswith("$x$"):  # It's a custom one with a full path
                replacement_source = {
                    "name": new_image[3:]
                }
            else:
                replacement_source = {
                    "name": f"remastered/field2d/us/{new_image}field.2dd/US_{new_image}field_2dd2.png",
                    "type": "internal"
                }
            # These images aren't any different between the regions, so save some mod space by using a multi
            assets.append({
                "platform": "pc",
                "name": f"remastered/field2d/us/{old_image}field.2dd/US_{old_image}field_2dd2.png",
                "multi": [
                    {"name": f"remastered/field2d/fr/{old_image}field.2dd/FR_{old_image}field_2dd2.png"},
                    {"name": f"remastered/field2d/gr/{old_image}field.2dd/GR_{old_image}field_2dd2.png"},
                    {"name": f"remastered/field2d/it/{old_image}field.2dd/IT_{old_image}field_2dd2.png"},
                    {"name": f"remastered/field2d/sp/{old_image}field.2dd/SP_{old_image}field_2dd2.png"}
                ],
                "method": "copy",
                "source": [replacement_source]
            })
        return assets

    def _compute_transition_replacements(self) -> dict[str, str]:
        supported_transitions = [
            AGRABAH,
            BEAST_CASTLE,
            PORT_ROYAL,
            DISNEY_CASTLE,
            TWTNW,
            HOLLOW_BASTION,
            OLYMPUS_COLISEUM,
            PRIDE_LANDS,
            ATLANTICA,
            LAND_OF_DRAGONS,
            HALLOWEEN_TOWN,
            HUNDRED_ACRE_WOOD,
            SPACE_PARANOIDS,
            TWILIGHT_TOWN,
            TIMELESS_RIVER
        ]

        transition_choice = self.transition_choice

        source_list: list[str] = []
        if transition_choice == configDict.RANDOMIZE_IN_GAME_ONLY:
            source_list = supported_transitions.copy()
        elif transition_choice == configDict.RANDOMIZE_CUSTOM_ONLY:
            source_list = [f"$x${path}" for path in self._custom_room_transition_images().values()]
        elif transition_choice == configDict.RANDOMIZE_ALL:
            source_list = []
            source_list.extend(supported_transitions)
            source_list.extend([f"$x${path}" for path in self._custom_room_transition_images().values()])

        transition_replacements: dict[str, str] = {}
        if len(source_list) == 0:
            return transition_replacements

        candidates: list[str] = []
        for index, old_transition in enumerate(supported_transitions):
            if len(candidates) == 0:
                candidates = source_list.copy()
                random.shuffle(candidates)
            transition_replacements[old_transition] = candidates.pop()

        return transition_replacements

    @staticmethod
    def _custom_room_transition_images() -> dict[str, Path]:
        result: dict[str, Path] = {}

        custom_visuals_path = appconfig.read_custom_visuals_path()
        if custom_visuals_path is None:
            return result

        room_transition_images_path = custom_visuals_path / "room-transition-images"
        if not room_transition_images_path.is_dir():
            return result

        for root, dirs, files in os.walk(room_transition_images_path):
            root_path = Path(root)
            for file in files:
                name, extension = os.path.splitext(file)
                if extension.lower() == ".png":
                    file_path = root_path / file
                    result[name] = file_path

        return result



from dataclasses import dataclass, field

from List.configDict import itemType
from List.inventory.item import InventoryItem


@dataclass(frozen=True)
class Shield(InventoryItem):
    id: int
    name: str
    type: itemType = field(init=False, default=itemType.SHIELD)
    stat_entry: int


AkashicRecord = Shield(146, "Akashic Record", stat_entry=107)
FrozenPridePlus = Shield(553, "Frozen Pride+", stat_entry=158)
GenjiShield = Shield(145, "Genji Shield", stat_entry=106)
MajesticMushroom = Shield(556, "Majestic Mushroom", stat_entry=161)
MajesticMushroomPlus = Shield(557, "Majestic Mushroom+", stat_entry=162)
NobodyGuard = Shield(147, "Nobody Guard", stat_entry=108)
OgreShield = Shield(141, "Ogre Shield", stat_entry=102)
SaveTheKingPlus = Shield(504, "Save The King+", stat_entry=147)
UltimateMushroom = Shield(558, "Ultimate Mushroom", stat_entry=163)



from Class.newLocationClass import KH2Location
from List.configDict import locationCategory, locationType


class LocationId:
    KingdomKey = 80
    Oathkeeper = 81
    Oblivion = 82
    AlphaWeapon = 83
    OmegaWeapon = 84
    Pureblood = 85
    KingdomKeyD = 116
    StruggleSword = 122
    StarSeeker = 123
    HiddenDragon = 124
    HerosCrest = 127
    Monochrome = 128
    FollowTheWind = 129
    CircleOfLife = 130
    PhotonDebugger = 131
    GullWing = 132
    RumblingRose = 133
    GuardianSoul = 134
    WishingLamp = 135
    DecisivePumpkin = 136
    SleepingLion = 137
    SweetMemories = 138
    MysteriousAbyss = 139
    FatalCrest = 140
    BondOfFlame = 141
    Fenrir = 142
    UltimaWeapon = 143
    StruggleWand = 144
    StruggleHammer = 145
    TwoBecomeOne = 148
    WinnersProof = 149

    HammerStaff = 87
    CometStaff = 90
    LordsBroom = 91
    MagesStaff = 86
    VictoryBell = 88
    MeteorStaff = 89
    WisdomWand = 92
    RisingDragon = 93
    NobodyLance = 94
    ShamansRelic = 95
    SaveTheQueenPlus = 146
    CenturionPlus = 151
    PreciousMushroom = 154
    PreciousMushroomPlus = 155
    PremiumMushroom = 156

    KnightsShield = 99
    AdamantShield = 100
    ChainGear = 101
    OgreShield = 102
    FallingStar = 103
    DreamCloud = 104
    KnightDefender = 105
    GenjiShield = 106
    AkashicRecord = 107
    NobodyGuard = 108
    SaveTheKingPlus = 147
    MajesticMushroom = 161
    FrozenPridePlus = 158
    MajesticMushroomPlus = 162
    UltimateMushroom = 163


class CheckLocation:
    KingdomKeyD = "Kingdom Key D (Slot)"
    AlphaWeapon = "Alpha Weapon (Slot)"
    OmegaWeapon = "Omega Weapon (Slot)"
    KingdomKey = "Kingdom Key (Slot)"
    Oathkeeper = "Oathkeeper (Slot)"
    Oblivion = "Oblivion (Slot)"
    StarSeeker = "Star Seeker (Slot)"
    HiddenDragon = "Hidden Dragon (Slot)"
    HerosCrest = "Hero's Crest (Slot)"
    Monochrome = "Monochrome (Slot)"
    FollowTheWind = "Follow the Wind (Slot)"
    CircleOfLife = "Circle of Life (Slot)"
    PhotonDebugger = "Photon Debugger (Slot)"
    GullWing = "Gull Wing (Slot)"
    RumblingRose = "Rumbling Rose (Slot)"
    GuardianSoul = "Guardian Soul (Slot)"
    WishingLamp = "Wishing Lamp (Slot)"
    DecisivePumpkin = "Decisive Pumpkin (Slot)"
    SweetMemories = "Sweet Memories (Slot)"
    MysteriousAbyss = "Mysterious Abyss (Slot)"
    SleepingLion = "Sleeping Lion (Slot)"
    BondOfFlame = "Bond of Flame (Slot)"
    TwoBecomeOne = "Two Become One (Slot)"
    FatalCrest = "Fatal Crest (Slot)"
    Fenrir = "Fenrir (Slot)"
    UltimaWeapon = "Ultima Weapon (Slot)"
    WinnersProof = "Winner's Proof (Slot)"
    Pureblood = "Pureblood (Slot)"

    StruggleSword = "Struggle Sword (Slot)"
    StruggleWand = "Struggle Wand (Slot)"
    StruggleHammer = "Struggle Hammer (Slot)"

    CenturionPlus = "Centurion+"
    CometStaff = "Comet Staff"
    HammerStaff = "Hammer Staff"
    LordsBroom = "Lord's Broom"
    MagesStaff = "Mage's Staff"
    MeteorStaff = "Meteor Staff"
    NobodyLance = "Nobody Lance"
    PreciousMushroom = "Precious Mushroom"
    PreciousMushroomPlus = "Precious Mushroom+"
    PremiumMushroom = "Premium Mushroom"
    RisingDragon = "Rising Dragon"
    SaveTheQueenPlus = "Save the Queen+"
    ShamansRelic = "Shaman's Relic"
    VictoryBell = "Victory Bell"
    WisdomWand = "Wisdom Wand"

    AdamantShield = "Adamant Shield"
    AkashicRecord = "Akashic Record"
    ChainGear = "Chain Gear"
    DreamCloud = "Dream Cloud"
    FallingStar = "Falling Star"
    FrozenPridePlus = "Frozen Pride+"
    GenjiShield = "Genji Shield"
    KnightDefender = "Knight Defender"
    KnightsShield = "Knight's Shield"
    MajesticMushroom = "Majestic Mushroom"
    MajesticMushroomPlus = "Majestic Mushroom+"
    NobodyGuard = "Nobody Guard"
    OgreShield = "Ogre Shield"
    SaveTheKingPlus = "Save The King+"
    UltimateMushroom = "Ultimate Mushroom"


def weapon_slot(loc_id: int, description: str) -> KH2Location:
    return KH2Location(loc_id, description, locationCategory.WEAPONSLOT, [locationType.WeaponSlot])


def keyblade_slots() -> list[KH2Location]:
    return [
        weapon_slot(LocationId.KingdomKeyD, CheckLocation.KingdomKeyD),
        weapon_slot(LocationId.AlphaWeapon, CheckLocation.AlphaWeapon),
        weapon_slot(LocationId.OmegaWeapon, CheckLocation.OmegaWeapon),
        weapon_slot(LocationId.KingdomKey, CheckLocation.KingdomKey),
        weapon_slot(LocationId.Oathkeeper, CheckLocation.Oathkeeper),
        weapon_slot(LocationId.Oblivion, CheckLocation.Oblivion),
        weapon_slot(LocationId.StarSeeker, CheckLocation.StarSeeker),
        weapon_slot(LocationId.HiddenDragon, CheckLocation.HiddenDragon),
        weapon_slot(LocationId.HerosCrest, CheckLocation.HerosCrest),
        weapon_slot(LocationId.Monochrome, CheckLocation.Monochrome),
        weapon_slot(LocationId.FollowTheWind, CheckLocation.FollowTheWind),
        weapon_slot(LocationId.CircleOfLife, CheckLocation.CircleOfLife),
        weapon_slot(LocationId.PhotonDebugger, CheckLocation.PhotonDebugger),
        weapon_slot(LocationId.GullWing, CheckLocation.GullWing),
        weapon_slot(LocationId.RumblingRose, CheckLocation.RumblingRose),
        weapon_slot(LocationId.GuardianSoul, CheckLocation.GuardianSoul),
        weapon_slot(LocationId.WishingLamp, CheckLocation.WishingLamp),
        weapon_slot(LocationId.DecisivePumpkin, CheckLocation.DecisivePumpkin),
        weapon_slot(LocationId.SweetMemories, CheckLocation.SweetMemories),
        weapon_slot(LocationId.MysteriousAbyss, CheckLocation.MysteriousAbyss),
        weapon_slot(LocationId.SleepingLion, CheckLocation.SleepingLion),
        weapon_slot(LocationId.BondOfFlame, CheckLocation.BondOfFlame),
        weapon_slot(LocationId.TwoBecomeOne, CheckLocation.TwoBecomeOne),
        weapon_slot(LocationId.FatalCrest, CheckLocation.FatalCrest),
        weapon_slot(LocationId.Fenrir, CheckLocation.Fenrir),
        weapon_slot(LocationId.UltimaWeapon, CheckLocation.UltimaWeapon),
        weapon_slot(LocationId.WinnersProof, CheckLocation.WinnersProof),
        weapon_slot(LocationId.Pureblood, CheckLocation.Pureblood),
    ]


def struggle_weapon_slots() -> list[KH2Location]:
    return [
        weapon_slot(LocationId.StruggleSword, CheckLocation.StruggleSword),
        weapon_slot(LocationId.StruggleWand, CheckLocation.StruggleWand),
        weapon_slot(LocationId.StruggleHammer, CheckLocation.StruggleHammer),
    ]


def donald_staff_slots() -> list[KH2Location]:
    return [
        weapon_slot(LocationId.CenturionPlus, CheckLocation.CenturionPlus),
        weapon_slot(LocationId.CometStaff, CheckLocation.CometStaff),
        weapon_slot(LocationId.HammerStaff, CheckLocation.HammerStaff),
        weapon_slot(LocationId.LordsBroom, CheckLocation.LordsBroom),
        weapon_slot(LocationId.MagesStaff, CheckLocation.MagesStaff),
        weapon_slot(LocationId.MeteorStaff, CheckLocation.MeteorStaff),
        weapon_slot(LocationId.NobodyLance, CheckLocation.NobodyLance),
        weapon_slot(LocationId.PreciousMushroom, CheckLocation.PreciousMushroom),
        weapon_slot(LocationId.PreciousMushroomPlus, CheckLocation.PreciousMushroomPlus),
        weapon_slot(LocationId.PremiumMushroom, CheckLocation.PremiumMushroom),
        weapon_slot(LocationId.RisingDragon, CheckLocation.RisingDragon),
        weapon_slot(LocationId.SaveTheQueenPlus, CheckLocation.SaveTheQueenPlus),
        weapon_slot(LocationId.ShamansRelic, CheckLocation.ShamansRelic),
        weapon_slot(LocationId.VictoryBell, CheckLocation.VictoryBell),
        weapon_slot(LocationId.WisdomWand, CheckLocation.WisdomWand),
    ]


def goofy_shield_slots() -> list[KH2Location]:
    return [
        weapon_slot(LocationId.AdamantShield, CheckLocation.AdamantShield),
        weapon_slot(LocationId.AkashicRecord, CheckLocation.AkashicRecord),
        weapon_slot(LocationId.ChainGear, CheckLocation.ChainGear),
        weapon_slot(LocationId.DreamCloud, CheckLocation.DreamCloud),
        weapon_slot(LocationId.FallingStar, CheckLocation.FallingStar),
        weapon_slot(LocationId.FrozenPridePlus, CheckLocation.FrozenPridePlus),
        weapon_slot(LocationId.GenjiShield, CheckLocation.GenjiShield),
        weapon_slot(LocationId.KnightDefender, CheckLocation.KnightDefender),
        weapon_slot(LocationId.KnightsShield, CheckLocation.KnightsShield),
        weapon_slot(LocationId.MajesticMushroom, CheckLocation.MajesticMushroom),
        weapon_slot(LocationId.MajesticMushroomPlus, CheckLocation.MajesticMushroomPlus),
        weapon_slot(LocationId.NobodyGuard, CheckLocation.NobodyGuard),
        weapon_slot(LocationId.OgreShield, CheckLocation.OgreShield),
        weapon_slot(LocationId.SaveTheKingPlus, CheckLocation.SaveTheKingPlus),
        weapon_slot(LocationId.UltimateMushroom, CheckLocation.UltimateMushroom),
    ]



from dataclasses import dataclass, field
from enum import Enum

from List.configDict import itemType
from List.inventory.item import InventoryItem


class GrowthType(Enum):
    HIGH_JUMP = "High Jump"
    QUICK_RUN = "Quick Run"
    AERIAL_DODGE = "Aerial Dodge"
    GLIDE = "Glide"
    DODGE_ROLL = "Dodge Roll"


@dataclass(frozen=True)
class GrowthAbility(InventoryItem):
    id: int
    name: str
    type: itemType = field(init=False, default=itemType.GROWTH_ABILITY)
    growth_type: GrowthType
    growth_level: int


HighJump1 = GrowthAbility(94, "High Jump", GrowthType.HIGH_JUMP, 1)
HighJump2 = GrowthAbility(95, "High Jump", GrowthType.HIGH_JUMP, 2)
HighJump3 = GrowthAbility(96, "High Jump", GrowthType.HIGH_JUMP, 3)
HighJumpMax = GrowthAbility(97, "High Jump", GrowthType.HIGH_JUMP, 4)
QuickRun1 = GrowthAbility(98, "Quick Run", GrowthType.QUICK_RUN, 1)
QuickRun2 = GrowthAbility(99, "Quick Run", GrowthType.QUICK_RUN, 2)
QuickRun3 = GrowthAbility(100, "Quick Run", GrowthType.QUICK_RUN, 3)
QuickRunMax = GrowthAbility(101, "Quick Run", GrowthType.QUICK_RUN, 4)
AerialDodge1 = GrowthAbility(102, "Aerial Dodge", GrowthType.AERIAL_DODGE, 1)
AerialDodge2 = GrowthAbility(103, "Aerial Dodge", GrowthType.AERIAL_DODGE, 2)
AerialDodge3 = GrowthAbility(104, "Aerial Dodge", GrowthType.AERIAL_DODGE, 3)
AerialDodgeMax = GrowthAbility(105, "Aerial Dodge", GrowthType.AERIAL_DODGE, 4)
Glide1 = GrowthAbility(106, "Glide", GrowthType.GLIDE, 1)
Glide2 = GrowthAbility(107, "Glide", GrowthType.GLIDE, 2)
Glide3 = GrowthAbility(108, "Glide", GrowthType.GLIDE, 3)
GlideMax = GrowthAbility(109, "Glide", GrowthType.GLIDE, 4)
DodgeRoll1 = GrowthAbility(564, "Dodge Roll", GrowthType.DODGE_ROLL, 1)
DodgeRoll2 = GrowthAbility(565, "Dodge Roll", GrowthType.DODGE_ROLL, 2)
DodgeRoll3 = GrowthAbility(566, "Dodge Roll", GrowthType.DODGE_ROLL, 3)
DodgeRollMax = GrowthAbility(567, "Dodge Roll", GrowthType.DODGE_ROLL, 4)


def all_growth() -> list[GrowthAbility]:
    return [
        HighJump1, HighJump2, HighJump3, HighJumpMax,
        QuickRun1, QuickRun2, QuickRun3, QuickRunMax,
        AerialDodge1, AerialDodge2, AerialDodge3, AerialDodgeMax,
        Glide1, Glide2, Glide3, GlideMax,
        DodgeRoll1, DodgeRoll2, DodgeRoll3, DodgeRollMax,
    ]


def all_growth_to_level(growth_level: int) -> list[GrowthAbility]:
    return [g for g in all_growth() if g.growth_level <= growth_level]


def growth_to_level(growth_level: int, growth_type: GrowthType) -> list[GrowthAbility]:
    return [g for g in all_growth_to_level(growth_level) if g.growth_type == growth_type]



import textwrap


SEED_SPLITTER = '$'


def loop(int_val):
    if int_val==47:
        return 122
    if int_val==58:
        return 65
    if int_val==64:
        return 57
    if int_val==91:
        return 97
    if int_val==96:
        return 90
    if int_val==123:
        return 48
    return int_val
    

def transform(in_string,tourney_gen):
    if not tourney_gen:
        return in_string
    out_string = ""
    for i in range(len(in_string)):
        out_string+=chr(loop((ord(in_string[i])+1)))
    
    return out_string


def untransform(in_string,tourney_gen):
    if not tourney_gen:
        return in_string
    out_string = ""
    for i in range(len(in_string)):
        out_string+=chr(loop((ord(in_string[i])-1)))
    
    return out_string

class SharedSeed:

    def __init__(self, generator_version: str, seed_name: str, spoiler_log: bool, settings_string: str, tourney_gen: bool = False):
        self.generator_version = generator_version
        self.seed_name = seed_name
        self.spoiler_log = spoiler_log
        self.settings_string = settings_string
        self.tourney_gen = tourney_gen

    def to_share_string(self) -> str:
        transformed_string = transform(self.seed_name,self.tourney_gen)

        return SEED_SPLITTER.join([
            self.generator_version,
            '1' if self.tourney_gen else '0',
            transformed_string,
            '1' if self.spoiler_log else '0',
            self.settings_string
        ])

    @classmethod
    def from_share_string(cls, local_generator_version: str, share_string: str):
        parts = share_string.split(SEED_SPLITTER)

        if len(parts) != 5:
            raise InvalidShareStringFormatException(textwrap.dedent(
                '''
                Unrecognized seed format.
                Make sure you are on the same seed generator version as the person who shared the seed with you.
                '''
            ))

        seed_generator_version = parts[0]
        if seed_generator_version != local_generator_version:
            raise IncompatibleShareStringVersionException(textwrap.dedent(
                '''
                Incompatible seed versions.
                Seed was generated from {0} and your version is {1}.
                Make sure you are on the same seed generator version as the person who shared the seed with you.
                '''.format(seed_generator_version, local_generator_version)
            ))

        tourney_gen = (parts[1] == '1')
        untransformed_string = untransform(parts[2],tourney_gen)

        return SharedSeed(
            generator_version=seed_generator_version,
            seed_name=untransformed_string,
            spoiler_log=True if parts[3] == '1' else False,
            settings_string=parts[4],
            tourney_gen=tourney_gen,
        )


class ShareStringException(Exception):

    def __init__(self, message):
        self.message = message


class InvalidShareStringFormatException(ShareStringException):

    def __init__(self, message):
        super().__init__(message)


class IncompatibleShareStringVersionException(ShareStringException):

    def __init__(self, message):
        super().__init__(message)



from dataclasses import dataclass, field
from typing import Optional

from List.configDict import itemType, locationCategory
from List.inventory.item import InventoryItem


@dataclass(frozen=True)
class DriveForm(InventoryItem):
    id: int
    name: str
    short_name: str
    level_category: Optional[locationCategory]
    type: itemType = field(init=False, default=itemType.FORM)


ValorForm = DriveForm(26, "Valor Form", "Valor", locationCategory.VALORLEVEL)
WisdomForm = DriveForm(27, "Wisdom Form", "Wisdom", locationCategory.WISDOMLEVEL)
FinalForm = DriveForm(29, "Final Form", "Final", locationCategory.FINALLEVEL)
AntiForm = DriveForm(30, "Anti-Form", "Anti", None)
MasterForm = DriveForm(31, "Master Form", "Master", locationCategory.MASTERLEVEL)
LimitForm = DriveForm(563, "Limit Form", "Limit", locationCategory.LIMITLEVEL)

DummyFinalForm = DriveForm(115, "Final Form", "Final", locationCategory.FINALLEVEL)  # (was Window of Time Map 2)
DummyValorForm = DriveForm(89, "Valor Form", "Valor", locationCategory.VALORLEVEL)  # (was Navigational Map)
